import os.path
import xml.etree.ElementTree as ET

import pytest

import salt.modules.virt as virt
import salt.syspaths
import salt.utils.xmlutil as xmlutil
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, patch

from .conftest import loader_modules_config
from .test_helpers import append_to_XMLDesc, assert_called, assert_equal_unit, strip_xml


@pytest.fixture
def configure_loader_modules():
    return loader_modules_config()


@pytest.mark.parametrize(
    "loader",
    [
        "/usr/lib/grub2/x86_64-xen/grub.xen",
        "/usr/share/grub2/x86_64-xen/grub.xen",
        None,
    ],
)
def test_gen_xml_for_xen_default_profile(loader, minion_opts):
    """
    Test virt._gen_xml(), XEN PV default profile case
    """
    diskp = virt._disk_profile(
        virt.libvirt.openAuth.return_value, "default", "xen", [], "hello"
    )
    nicp = virt._nic_profile("default", "xen")
    with patch.dict(
        virt.__grains__, {"os_family": "Suse"}  # pylint: disable=no-member
    ):
        os_mock = MagicMock(spec=virt.os)

        def fake_exists(path):
            return loader and path == loader

        os_mock.path.exists = MagicMock(side_effect=fake_exists)

        with patch.dict(virt.__dict__, {"os": os_mock}):
            if loader:
                xml_data = virt._gen_xml(
                    virt.libvirt.openAuth.return_value,
                    "hello",
                    1,
                    512,
                    diskp,
                    nicp,
                    "xen",
                    "xen",
                    "x86_64",
                    boot=None,
                )
                root = ET.fromstring(xml_data)
                assert root.attrib["type"] == "xen"
                assert root.find("vcpu").text == "1"
                assert root.find("memory").text == str(512 * 1024)
                assert root.find("memory").attrib["unit"] == "KiB"
                assert root.find(".//kernel").text == loader

                disks = root.findall(".//disk")
                assert len(disks) == 1
                disk = disks[0]
                root_dir = salt.syspaths.ROOT_DIR
                assert disk.find("source").attrib["file"].startswith(root_dir)
                assert "hello_system" in disk.find("source").attrib["file"]
                assert disk.find("target").attrib["dev"] == "xvda"
                assert disk.find("target").attrib["bus"] == "xen"
                assert disk.find("driver").attrib["name"] == "qemu"
                assert disk.find("driver").attrib["type"] == "qcow2"

                interfaces = root.findall(".//interface")
                assert len(interfaces) == 1
                iface = interfaces[0]
                assert iface.attrib["type"] == "bridge"
                assert iface.find("source").attrib["bridge"] == "br0"
                assert iface.find("model") is None
            else:
                with pytest.raises(CommandExecutionError):
                    xml_data = virt._gen_xml(
                        virt.libvirt.openAuth.return_value,
                        "hello",
                        1,
                        512,
                        diskp,
                        nicp,
                        "xen",
                        "xen",
                        "x86_64",
                        boot=None,
                    )


def test_update_xen_disk_volumes(make_mock_vm, make_mock_storage_pool):
    xml_def = """
        <domain type='xen'>
          <name>my_vm</name>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <os>
            <type arch='x86_64'>linux</type>
            <kernel>/usr/lib/grub2/x86_64-xen/grub.xen</kernel>
          </os>
          <devices>
            <disk type='file' device='disk'>
              <driver name='qemu' type='qcow2' cache='none' io='native'/>
              <source file='/path/to/default/my_vm_system'/>
              <target dev='xvda' bus='xen'/>
            </disk>
            <disk type='block' device='disk'>
              <driver name='qemu' type='raw' cache='none' io='native'/>
              <source dev='/path/to/my-iscsi/unit:0:0:1'/>
              <target dev='xvdb' bus='xen'/>
            </disk>
            <controller type='xenbus' index='0'/>
          </devices>
        </domain>"""
    domain_mock = make_mock_vm(xml_def)
    make_mock_storage_pool("default", "dir", ["my_vm_system"])
    make_mock_storage_pool("my-iscsi", "iscsi", ["unit:0:0:1"])
    make_mock_storage_pool("vdb", "disk", ["vdb1"])

    ret = virt.update(
        "my_vm",
        disks=[
            {"name": "system", "pool": "default"},
            {"name": "iscsi-data", "pool": "my-iscsi", "source_file": "unit:0:0:1"},
            {"name": "vdb-data", "pool": "vdb", "source_file": "vdb1"},
            {"name": "file-data", "pool": "default", "size": "10240"},
        ],
    )

    assert ret["definition"]
    virt.libvirt.openAuth().defineXML = virt.libvirt.openAuth().defineXML
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find(".//disk[3]").get("type") == "block"
    assert setxml.find(".//disk[3]/source").get("dev") == "/path/to/vdb/vdb1"

    # Note that my_vm-file-data was not an existing volume before the update
    assert setxml.find(".//disk[4]").get("type") == "file"
    assert (
        setxml.find(".//disk[4]/source").get("file")
        == "/path/to/default/my_vm_file-data"
    )


def test_get_disks(make_mock_vm, make_mock_storage_pool):
    # test with volumes
    vm_def = """<domain type='kvm' id='3'>
      <name>srv01</name>
      <devices>
        <disk type='volume' device='disk'>
          <driver name='qemu' type='qcow2' cache='none' io='native'/>
          <source pool='default' volume='srv01_system'/>
          <backingStore/>
          <target dev='vda' bus='virtio'/>
          <alias name='virtio-disk0'/>
          <address type='pci' domain='0x0000' bus='0x00' slot='0x04' function='0x0'/>
        </disk>
        <disk type='volume' device='disk'>
          <driver name='qemu' type='qcow2' cache='none' io='native'/>
          <source pool='default' volume='srv01_data'/>
          <backingStore type='file' index='1'>
            <format type='qcow2'/>
            <source file='/var/lib/libvirt/images/vol01'/>
            <backingStore/>
          </backingStore>
          <target dev='vdb' bus='virtio'/>
          <alias name='virtio-disk1'/>
          <address type='pci' domain='0x0000' bus='0x00' slot='0x05' function='0x0'/>
        </disk>
        <disk type='volume' device='disk'>
          <driver name='qemu' type='qcow2' cache='none' io='native'/>
          <source pool='default' volume='vm05_system'/>
          <backingStore type='file' index='1'>
            <format type='qcow2'/>
            <source file='/var/lib/libvirt/images/vm04_system.qcow2'/>
            <backingStore type='file' index='2'>
              <format type='raw'/>
              <source file='/var/testsuite-data/disk-image-template.raw'/>
              <backingStore/>
            </backingStore>
          </backingStore>
          <target dev='vdc' bus='virtio'/>
          <alias name='virtio-disk0'/>
          <address type='pci' domain='0x0000' bus='0x00' slot='0x04' function='0x0'/>
        </disk>
        <disk type='volume' device='disk'>
          <driver name='qemu' type='qcow2' cache='none' io='native'/>
          <source pool='stopped' volume='vm05_data'/>
          <target dev='vdd' bus='virtio'/>
        </disk>
        <disk type='network' device='cdrom'>
          <driver name='qemu' type='raw' cache='none' io='native'/>
          <source protocol='http' name='/pub/iso/myimage.iso' query='foo=bar&amp;baz=flurb' index='1'>
            <host name='dev-srv.tf.local' port='80'/>
          </source>
          <target dev='hda' bus='ide'/>
          <readonly/>
          <alias name='ide0-0-0'/>
          <address type='drive' controller='0' bus='0' target='0' unit='0'/>
        </disk>
      </devices>
    </domain>
    """
    make_mock_vm(vm_def)

    pool_mock = make_mock_storage_pool(
        "default", "dir", ["srv01_system", "srv01_data", "vm05_system"]
    )
    make_mock_storage_pool("stopped", "dir", [])

    # Append backing store to srv01_data volume XML description
    srv1data_mock = pool_mock.storageVolLookupByName("srv01_data")
    append_to_XMLDesc(
        srv1data_mock,
        """
        <backingStore>
          <path>/var/lib/libvirt/images/vol01</path>
          <format type="qcow2"/>
        </backingStore>""",
    )

    assert virt.get_disks("srv01") == {
        "vda": {
            "type": "disk",
            "file": "default/srv01_system",
            "file format": "qcow2",
            "disk size": 12345,
            "virtual size": 1234567,
        },
        "vdb": {
            "type": "disk",
            "file": "default/srv01_data",
            "file format": "qcow2",
            "disk size": 12345,
            "virtual size": 1234567,
            "backing file": {
                "file": "/var/lib/libvirt/images/vol01",
                "file format": "qcow2",
            },
        },
        "vdc": {
            "type": "disk",
            "file": "default/vm05_system",
            "file format": "qcow2",
            "disk size": 12345,
            "virtual size": 1234567,
            "backing file": {
                "file": "/var/lib/libvirt/images/vm04_system.qcow2",
                "file format": "qcow2",
                "backing file": {
                    "file": "/var/testsuite-data/disk-image-template.raw",
                    "file format": "raw",
                },
            },
        },
        "vdd": {"type": "disk", "file": "stopped/vm05_data", "file format": "qcow2"},
        "hda": {
            "type": "cdrom",
            "file format": "raw",
            "file": "http://dev-srv.tf.local:80/pub/iso/myimage.iso?foo=bar&baz=flurb",
        },
    }


def test_get_disk_convert_volumes(make_mock_vm, make_mock_storage_pool):
    vm_def = """<domain type='kvm' id='3'>
      <name>srv01</name>
      <devices>
        <disk type='file' device='disk'>
          <driver name='qemu' type='qcow2' cache='none' io='native'/>
          <source file='/path/to/default/srv01_system'/>
          <target dev='vda' bus='virtio'/>
          <alias name='virtio-disk0'/>
          <address type='pci' domain='0x0000' bus='0x00' slot='0x04' function='0x0'/>
        </disk>
        <disk type='block' device='disk'>
          <driver name='qemu' type='raw'/>
          <source dev='/path/to/default/srv01_data'/>
          <target dev='vdb' bus='virtio'/>
          <alias name='virtio-disk1'/>
          <address type='pci' domain='0x0000' bus='0x00' slot='0x05' function='0x0'/>
        </disk>
        <disk type='file' device='disk'>
          <driver name='qemu' type='qcow2' cache='none' io='native'/>
          <source file='/path/to/srv01_extra'/>
          <target dev='vdc' bus='virtio'/>
          <alias name='virtio-disk1'/>
          <address type='pci' domain='0x0000' bus='0x00' slot='0x05' function='0x0'/>
        </disk>
      </devices>
    </domain>
    """
    domain_mock = make_mock_vm(vm_def)

    pool_mock = make_mock_storage_pool("default", "dir", ["srv01_system", "srv01_data"])

    subprocess_mock = MagicMock()
    popen_mock = MagicMock(spec=virt.subprocess.Popen)
    popen_mock.return_value.communicate.return_value = [
        """[
        {
            "virtual-size": 214748364800,
            "filename": "/path/to/srv01_extra",
            "cluster-size": 65536,
            "format": "qcow2",
            "actual-size": 340525056,
            "format-specific": {
                "type": "qcow2",
                "data": {
                    "compat": "1.1",
                    "lazy-refcounts": false,
                    "refcount-bits": 16,
                    "corrupt": false
                }
            },
            "dirty-flag": false
        }
    ]
    """,
        "",
    ]
    popen_mock.return_value.returncode = 0
    subprocess_mock.Popen = popen_mock

    with patch.dict(virt.__dict__, {"subprocess": subprocess_mock}):
        assert virt.get_disks("srv01") == {
            "vda": {
                "type": "disk",
                "file": "default/srv01_system",
                "file format": "qcow2",
                "disk size": 12345,
                "virtual size": 1234567,
            },
            "vdb": {
                "type": "disk",
                "file": "default/srv01_data",
                "file format": "raw",
                "disk size": 12345,
                "virtual size": 1234567,
            },
            "vdc": {
                "type": "disk",
                "file": "/path/to/srv01_extra",
                "file format": "qcow2",
                "cluster size": 65536,
                "disk size": 340525056,
                "virtual size": 214748364800,
            },
        }


def test_get_disk_missing(make_mock_vm):
    """
    Test virt.get_disks when the file doesn't exist
    """
    vm_def = """<domain type='kvm' id='3'>
      <name>srv01</name>
      <devices>
        <disk type='file' device='disk'>
          <driver name='qemu' type='qcow2' cache='none' io='native'/>
          <source file='/path/to/default/srv01_system'/>
          <target dev='vda' bus='virtio'/>
        </disk>
      </devices>
    </domain>
    """
    domain_mock = make_mock_vm(vm_def)

    subprocess_mock = MagicMock()
    popen_mock = MagicMock(spec=virt.subprocess.Popen)
    popen_mock.return_value.communicate.return_value = ("", "File not found")
    popen_mock.return_value.returncode = 1
    subprocess_mock.Popen = popen_mock

    with patch.dict(virt.__dict__, {"subprocess": subprocess_mock}):
        assert virt.get_disks("srv01") == {
            "vda": {
                "type": "disk",
                "file": "/path/to/default/srv01_system",
                "file format": "qcow2",
                "error": "File not found",
            },
        }


def test_get_disk_no_qemuimg(make_mock_vm):
    """
    Test virt.get_disks when qemu_img can't be found
    """
    vm_def = """<domain type='kvm' id='3'>
      <name>srv01</name>
      <devices>
        <disk type='file' device='disk'>
          <driver name='qemu' type='qcow2' cache='none' io='native'/>
          <source file='/path/to/default/srv01_system'/>
          <target dev='vda' bus='virtio'/>
        </disk>
      </devices>
    </domain>
    """
    domain_mock = make_mock_vm(vm_def)

    subprocess_mock = MagicMock()
    subprocess_mock.Popen = MagicMock(
        side_effect=FileNotFoundError("No such file or directory: 'qemu-img'")
    )

    with patch.dict(virt.__dict__, {"subprocess": subprocess_mock}):
        assert virt.get_disks("srv01") == {
            "vda": {
                "type": "disk",
                "file": "/path/to/default/srv01_system",
                "file format": "qcow2",
                "error": "qemu-img not found",
            },
        }


def test_update_approx_mem(make_mock_vm):
    """
    test virt.update with memory parameter unchanged thought not exactly equals to the current value.
    This may happen since libvirt sometimes rounds the memory value.
    """
    xml_def = """
        <domain type="kvm">
          <name>my_vm</name>
          <memory unit='KiB'>3177680</memory>
          <currentMemory unit='KiB'>3177680</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <os>
            <type arch='x86_64'>hvm</type>
          </os>
          <on_reboot>restart</on_reboot>
        </domain>
    """
    domain_mock = make_mock_vm(xml_def)

    ret = virt.update("my_vm", mem={"boot": "3253941043B", "current": "3253941043B"})
    assert not ret["definition"]


def test_gen_hypervisor_features():
    """
    Test the virt._gen_xml hypervisor_features handling
    """
    xml_data = virt._gen_xml(
        virt.libvirt.openAuth.return_value,
        "hello",
        1,
        512,
        {},
        {},
        "kvm",
        "hvm",
        "x86_64",
        hypervisor_features={"kvm-hint-dedicated": True},
    )
    root = ET.fromstring(xml_data)
    assert root.find("features/kvm/hint-dedicated").attrib["state"] == "on"


def test_update_hypervisor_features(make_mock_vm):
    """
    Test changing the hypervisor features of a guest
    """
    xml_def = """
        <domain type="kvm">
          <name>my_vm</name>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <os>
            <type arch='x86_64'>linux</type>
            <kernel>/usr/lib/grub2/x86_64-xen/grub.xen</kernel>
          </os>
          <features>
            <kvm>
              <hint-dedicated state="on"/>
            </kvm>
          </features>
          <on_reboot>restart</on_reboot>
        </domain>
    """
    domain_mock = make_mock_vm(xml_def)

    # Update with no change to the features
    ret = virt.update("my_vm", hypervisor_features={"kvm-hint-dedicated": True})
    assert not ret["definition"]

    # Alter the features
    ret = virt.update("my_vm", hypervisor_features={"kvm-hint-dedicated": False})
    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("features/kvm/hint-dedicated").get("state") == "off"

    # Add the features
    xml_def = """
        <domain type="kvm">
          <name>my_vm</name>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <os>
            <type arch='x86_64'>linux</type>
            <kernel>/usr/lib/grub2/x86_64-xen/grub.xen</kernel>
          </os>
        </domain>
    """
    domain_mock = make_mock_vm(xml_def)
    ret = virt.update("my_vm", hypervisor_features={"kvm-hint-dedicated": True})
    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("features/kvm/hint-dedicated").get("state") == "on"


def test_gen_clock():
    """
    Test the virt._gen_xml clock property
    """
    # Localtime with adjustment
    xml_data = virt._gen_xml(
        virt.libvirt.openAuth.return_value,
        "hello",
        1,
        512,
        {},
        {},
        "kvm",
        "hvm",
        "x86_64",
        clock={"adjustment": 3600, "utc": False},
    )
    root = ET.fromstring(xml_data)
    assert root.find("clock").get("offset") == "localtime"
    assert root.find("clock").get("adjustment") == "3600"

    # Specific timezone
    xml_data = virt._gen_xml(
        virt.libvirt.openAuth.return_value,
        "hello",
        1,
        512,
        {},
        {},
        "kvm",
        "hvm",
        "x86_64",
        clock={"timezone": "CEST"},
    )
    root = ET.fromstring(xml_data)
    assert root.find("clock").get("offset") == "timezone"
    assert root.find("clock").get("timezone") == "CEST"

    # UTC
    xml_data = virt._gen_xml(
        virt.libvirt.openAuth.return_value,
        "hello",
        1,
        512,
        {},
        {},
        "kvm",
        "hvm",
        "x86_64",
        clock={"utc": True},
    )
    root = ET.fromstring(xml_data)
    assert root.find("clock").get("offset") == "utc"

    # Timers
    xml_data = virt._gen_xml(
        virt.libvirt.openAuth.return_value,
        "hello",
        1,
        512,
        {},
        {},
        "kvm",
        "hvm",
        "x86_64",
        clock={
            "timers": {
                "tsc": {"frequency": 3504000000, "mode": "native"},
                "rtc": {
                    "tickpolicy": "catchup",
                    "slew": 4636,
                    "threshold": 123,
                    "limit": 2342,
                },
                "hpet": {"present": False},
            },
        },
    )
    root = ET.fromstring(xml_data)
    assert root.find("clock").get("offset") == "utc"
    assert root.find("clock/timer[@name='tsc']").get("frequency") == "3504000000"
    assert root.find("clock/timer[@name='tsc']").get("mode") == "native"
    assert root.find("clock/timer[@name='rtc']").get("tickpolicy") == "catchup"
    assert root.find("clock/timer[@name='rtc']/catchup").attrib == {
        "slew": "4636",
        "threshold": "123",
        "limit": "2342",
    }
    assert root.find("clock/timer[@name='hpet']").get("present") == "no"


def test_update_clock(make_mock_vm):
    """
    test virt.update with clock parameter
    """
    xml_def = """
        <domain type="kvm">
          <name>my_vm</name>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <os>
            <type arch='x86_64'>linux</type>
            <kernel>/usr/lib/grub2/x86_64-xen/grub.xen</kernel>
          </os>
          <clock offset="localtime" adjustment="-3600">
            <timer name="tsc" frequency="3504000000" mode="native" />
            <timer name="kvmclock" present="no" />
          </clock>
          <on_reboot>restart</on_reboot>
        </domain>
    """
    domain_mock = make_mock_vm(xml_def)

    # Update with no change to the features
    ret = virt.update(
        "my_vm",
        clock={
            "utc": False,
            "adjustment": -3600,
            "timers": {
                "tsc": {"frequency": 3504000000, "mode": "native"},
                "kvmclock": {"present": False},
            },
        },
    )
    assert not ret["definition"]

    # Update
    ret = virt.update(
        "my_vm",
        clock={
            "timezone": "CEST",
            "timers": {
                "rtc": {
                    "track": "wall",
                    "tickpolicy": "catchup",
                    "slew": 4636,
                    "threshold": 123,
                    "limit": 2342,
                },
                "hpet": {"present": True},
            },
        },
    )
    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("clock").get("offset") == "timezone"
    assert setxml.find("clock").get("timezone") == "CEST"
    assert {t.get("name") for t in setxml.findall("clock/timer")} == {"rtc", "hpet"}
    assert setxml.find("clock/timer[@name='rtc']").get("tickpolicy") == "catchup"
    assert setxml.find("clock/timer[@name='rtc']").get("track") == "wall"
    assert setxml.find("clock/timer[@name='rtc']/catchup").attrib == {
        "slew": "4636",
        "threshold": "123",
        "limit": "2342",
    }
    assert setxml.find("clock/timer[@name='hpet']").get("present") == "yes"

    # Revert to UTC
    ret = virt.update("my_vm", clock={"utc": True, "adjustment": None, "timers": None})
    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("clock").attrib == {"offset": "utc"}
    assert setxml.find("clock/timer") is None


def test_update_stop_on_reboot_reset(make_mock_vm):
    """
    Test virt.update to remove the on_reboot=destroy flag
    """
    xml_def = """
        <domain type='kvm'>
          <name>my_vm</name>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <on_reboot>destroy</on_reboot>
          <os>
            <type arch='x86_64'>hvm</type>
          </os>
        </domain>"""
    domain_mock = make_mock_vm(xml_def)

    ret = virt.update("my_vm")

    assert ret["definition"]
    virt.libvirt.openAuth().defineXML = virt.libvirt.openAuth().defineXML
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("./on_reboot").text == "restart"


def test_update_stop_on_reboot(make_mock_vm):
    """
    Test virt.update to add the on_reboot=destroy flag
    """
    xml_def = """
        <domain type='kvm'>
          <name>my_vm</name>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <os>
            <type arch='x86_64'>hvm</type>
          </os>
        </domain>"""
    domain_mock = make_mock_vm(xml_def)

    ret = virt.update("my_vm", stop_on_reboot=True)

    assert ret["definition"]
    virt.libvirt.openAuth().defineXML = virt.libvirt.openAuth().defineXML
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("./on_reboot").text == "destroy"


def test_init_no_stop_on_reboot(make_capabilities):
    """
    Test virt.init to add the on_reboot=restart flag
    """
    make_capabilities()
    with patch.dict(virt.os.__dict__, {"chmod": MagicMock(), "makedirs": MagicMock()}):
        with patch.dict(virt.__salt__, {"cmd.run": MagicMock()}):
            virt.init("test_vm", 2, 2048, start=False)
            virt.libvirt.openAuth().defineXML = virt.libvirt.openAuth().defineXML
            setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
            assert setxml.find("./on_reboot").text == "restart"


def test_init_stop_on_reboot(make_capabilities):
    """
    Test virt.init to add the on_reboot=destroy flag
    """
    make_capabilities()
    with patch.dict(virt.os.__dict__, {"chmod": MagicMock(), "makedirs": MagicMock()}):
        with patch.dict(virt.__salt__, {"cmd.run": MagicMock()}):
            virt.init("test_vm", 2, 2048, stop_on_reboot=True, start=False)
            virt.libvirt.openAuth().defineXML = virt.libvirt.openAuth().defineXML
            setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
            assert setxml.find("./on_reboot").text == "destroy"


def test_init_hostdev_usb(make_capabilities, make_mock_device):
    """
    Test virt.init with USB host device passed through
    """
    make_capabilities()
    make_mock_device(
        """
        <device>
          <name>usb_3_1_3</name>
          <path>/sys/devices/pci0000:00/0000:00:1d.6/0000:06:00.0/0000:07:02.0/0000:3e:00.0/usb3/3-1/3-1.3</path>
          <devnode type='dev'>/dev/bus/usb/003/004</devnode>
          <parent>usb_3_1</parent>
          <driver>
            <name>usb</name>
          </driver>
          <capability type='usb_device'>
            <bus>3</bus>
            <device>4</device>
            <product id='0x6006'>AUKEY PC-LM1E Camera</product>
            <vendor id='0x0458'>KYE Systems Corp. (Mouse Systems)</vendor>
          </capability>
        </device>
    """
    )
    with patch.dict(virt.os.__dict__, {"chmod": MagicMock(), "makedirs": MagicMock()}):
        with patch.dict(virt.__salt__, {"cmd.run": MagicMock()}):
            virt.init("test_vm", 2, 2048, host_devices=["usb_3_1_3"], start=False)
            define_mock = virt.libvirt.openAuth().defineXML
            setxml = ET.fromstring(define_mock.call_args[0][0])
            expected_xml = strip_xml(
                """
                <hostdev mode='subsystem' type='usb'>
                  <source>
                    <vendor id='0x0458'/>
                    <product id='0x6006'/>
                  </source>
                </hostdev>
                """
            )
            assert (
                strip_xml(ET.tostring(setxml.find("./devices/hostdev"))) == expected_xml
            )


def test_init_hostdev_pci(make_capabilities, make_mock_device):
    """
    Test virt.init with PCI host device passed through
    """
    make_capabilities()
    make_mock_device(
        """
        <device>
          <name>pci_1002_71c4</name>
          <parent>pci_8086_27a1</parent>
          <capability type='pci'>
            <class>0xffffff</class>
            <domain>0</domain>
            <bus>1</bus>
            <slot>0</slot>
            <function>0</function>
            <product id='0x71c4'>M56GL [Mobility FireGL V5200]</product>
            <vendor id='0x1002'>ATI Technologies Inc</vendor>
            <numa node='1'/>
          </capability>
        </device>
    """
    )
    with patch.dict(virt.os.__dict__, {"chmod": MagicMock(), "makedirs": MagicMock()}):
        with patch.dict(virt.__salt__, {"cmd.run": MagicMock()}):
            virt.init("test_vm", 2, 2048, host_devices=["pci_1002_71c4"], start=False)
            define_mock = virt.libvirt.openAuth().defineXML
            setxml = ET.fromstring(define_mock.call_args[0][0])
            expected_xml = strip_xml(
                """
                <hostdev mode='subsystem' type='pci' managed='yes'>
                  <source>
                    <address domain='0x0000' bus='0x01' slot='0x00' function='0x0'/>
                  </source>
                </hostdev>
                """
            )
            assert (
                strip_xml(ET.tostring(setxml.find("./devices/hostdev"))) == expected_xml
            )


def test_update_hostdev_nochange(make_mock_device, make_mock_vm):
    """
    Test the virt.update function with no host device changes
    """
    xml_def = """
        <domain type='kvm'>
          <name>my_vm</name>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <os>
            <type arch='x86_64'>hvm</type>
          </os>
          <on_reboot>restart</on_reboot>
          <devices>
            <hostdev mode='subsystem' type='pci' managed='yes'>
              <source>
                <address domain='0x0000' bus='0x01' slot='0x00' function='0x0'/>
              </source>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
            </hostdev>
            <hostdev mode='subsystem' type='usb' managed='no'>
              <source>
                <vendor id='0x0458'/>
                <product id='0x6006'/>
                <address bus='3' device='4'/>
              </source>
              <alias name='hostdev0'/>
              <address type='usb' bus='0' port='1'/>
            </hostdev>
          </devices>
        </domain>"""
    domain_mock = make_mock_vm(xml_def)

    make_mock_device(
        """
        <device>
          <name>usb_3_1_3</name>
          <path>/sys/devices/pci0000:00/0000:00:1d.6/0000:06:00.0/0000:07:02.0/0000:3e:00.0/usb3/3-1/3-1.3</path>
          <devnode type='dev'>/dev/bus/usb/003/004</devnode>
          <parent>usb_3_1</parent>
          <driver>
            <name>usb</name>
          </driver>
          <capability type='usb_device'>
            <bus>3</bus>
            <device>4</device>
            <product id='0x6006'>AUKEY PC-LM1E Camera</product>
            <vendor id='0x0458'>KYE Systems Corp. (Mouse Systems)</vendor>
          </capability>
        </device>
    """
    )
    make_mock_device(
        """
        <device>
          <name>pci_1002_71c4</name>
          <parent>pci_8086_27a1</parent>
          <capability type='pci'>
            <class>0xffffff</class>
            <domain>0</domain>
            <bus>1</bus>
            <slot>0</slot>
            <function>0</function>
            <product id='0x71c4'>M56GL [Mobility FireGL V5200]</product>
            <vendor id='0x1002'>ATI Technologies Inc</vendor>
            <numa node='1'/>
          </capability>
        </device>
    """
    )

    ret = virt.update("my_vm", host_devices=["pci_1002_71c4", "usb_3_1_3"])

    assert not ret["definition"]
    define_mock = virt.libvirt.openAuth().defineXML
    define_mock.assert_not_called()


@pytest.mark.parametrize(
    "running,live",
    [(False, False), (True, False), (True, True)],
    ids=["stopped, no live", "running, no live", "running, live"],
)
def test_update_hostdev_changes(running, live, make_mock_device, make_mock_vm, test):
    """
    Test the virt.update function with host device changes
    """
    xml_def = """
        <domain type='kvm'>
          <name>my_vm</name>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <os>
            <type arch='x86_64'>hvm</type>
          </os>
          <on_reboot>restart</on_reboot>
          <devices>
            <hostdev mode='subsystem' type='pci' managed='yes'>
              <source>
                <address domain='0x0000' bus='0x01' slot='0x00' function='0x0'/>
              </source>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
            </hostdev>
          </devices>
        </domain>"""
    domain_mock = make_mock_vm(xml_def, running)

    make_mock_device(
        """
        <device>
          <name>usb_3_1_3</name>
          <path>/sys/devices/pci0000:00/0000:00:1d.6/0000:06:00.0/0000:07:02.0/0000:3e:00.0/usb3/3-1/3-1.3</path>
          <devnode type='dev'>/dev/bus/usb/003/004</devnode>
          <parent>usb_3_1</parent>
          <driver>
            <name>usb</name>
          </driver>
          <capability type='usb_device'>
            <bus>3</bus>
            <device>4</device>
            <product id='0x6006'>AUKEY PC-LM1E Camera</product>
            <vendor id='0x0458'>KYE Systems Corp. (Mouse Systems)</vendor>
          </capability>
        </device>
    """
    )

    make_mock_device(
        """
            <device>
              <name>pci_1002_71c4</name>
              <parent>pci_8086_27a1</parent>
              <capability type='pci'>
                <class>0xffffff</class>
                <domain>0</domain>
                <bus>1</bus>
                <slot>0</slot>
                <function>0</function>
                <product id='0x71c4'>M56GL [Mobility FireGL V5200]</product>
                <vendor id='0x1002'>ATI Technologies Inc</vendor>
                <numa node='1'/>
              </capability>
            </device>
        """
    )

    ret = virt.update("my_vm", host_devices=["usb_3_1_3"], test=test, live=live)
    define_mock = virt.libvirt.openAuth().defineXML
    assert_called(define_mock, not test)

    # Test that the XML is updated with the proper devices
    usb_device_xml = strip_xml(
        """
        <hostdev mode="subsystem" type="usb">
          <source>
           <vendor id="0x0458" />
           <product id="0x6006" />
          </source>
        </hostdev>
        """
    )
    if not test:
        set_xml = ET.fromstring(define_mock.call_args[0][0])
        actual_hostdevs = [
            ET.tostring(xmlutil.strip_spaces(node))
            for node in set_xml.findall("./devices/hostdev")
        ]
        assert actual_hostdevs == [usb_device_xml]

    if not test and live:
        attach_xml = strip_xml(domain_mock.attachDevice.call_args[0][0])
        assert attach_xml == usb_device_xml

        pci_device_xml = strip_xml(
            """
                <hostdev mode='subsystem' type='pci' managed='yes'>
                  <source>
                    <address domain='0x0000' bus='0x01' slot='0x00' function='0x0'/>
                  </source>
                  <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
                </hostdev>
            """
        )
        detach_xml = strip_xml(domain_mock.detachDevice.call_args[0][0])
        assert detach_xml == pci_device_xml
    else:
        domain_mock.attachDevice.assert_not_called()
        domain_mock.detachDevice.assert_not_called()


def test_diff_nics():
    """
    Test virt._diff_nics()
    """
    old_nics = ET.fromstring(
        """
        <devices>
           <interface type='network'>
             <mac address='52:54:00:39:02:b1'/>
             <source network='default'/>
             <model type='virtio'/>
             <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
           </interface>
           <interface type='network'>
             <mac address='52:54:00:39:02:b2'/>
             <source network='admin'/>
             <model type='virtio'/>
             <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
           </interface>
           <interface type='network'>
             <mac address='52:54:00:39:02:b3'/>
             <source network='admin'/>
             <model type='virtio'/>
             <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
           </interface>
        </devices>
    """
    ).findall("interface")

    new_nics = ET.fromstring(
        """
        <devices>
           <interface type='network'>
             <mac address='52:54:00:39:02:b1'/>
             <source network='default'/>
             <model type='virtio'/>
           </interface>
           <interface type='network'>
             <mac address='52:54:00:39:02:b2'/>
             <source network='default'/>
             <model type='virtio'/>
           </interface>
           <interface type='network'>
             <mac address='52:54:00:39:02:b4'/>
             <source network='admin'/>
             <model type='virtio'/>
           </interface>
        </devices>
    """
    ).findall("interface")
    ret = virt._diff_interface_lists(old_nics, new_nics)
    assert [nic.find("mac").get("address") for nic in ret["unchanged"]] == [
        "52:54:00:39:02:b1"
    ]
    assert [nic.find("mac").get("address") for nic in ret["new"]] == [
        "52:54:00:39:02:b2",
        "52:54:00:39:02:b4",
    ]
    assert [nic.find("mac").get("address") for nic in ret["deleted"]] == [
        "52:54:00:39:02:b2",
        "52:54:00:39:02:b3",
    ]


def test_diff_nics_live_nochange():
    """
    Libvirt alters the NICs of network type when running the guest, test the virt._diff_nics()
    function with no change in such a case.
    """
    old_nics = ET.fromstring(
        """
        <devices>
          <interface type='direct'>
            <mac address='52:54:00:03:02:15'/>
            <source network='test-vepa' portid='8377df4f-7c72-45f3-9ba4-a76306333396' dev='eth1' mode='vepa'/>
            <target dev='macvtap0'/>
            <model type='virtio'/>
            <alias name='net0'/>
            <address type='pci' domain='0x0000' bus='0x00' slot='0x05' function='0x0'/>
          </interface>
          <interface type='bridge'>
            <mac address='52:54:00:ea:2e:89'/>
            <source network='default' portid='b97ec5b7-25fd-4697-ae45-06af8cc1a964' bridge='br0'/>
            <target dev='vnet0'/>
            <model type='virtio'/>
            <alias name='net0'/>
            <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
          </interface>
        </devices>
        """
    ).findall("interface")

    new_nics = ET.fromstring(
        """
        <devices>
           <interface type='network'>
             <source network='test-vepa'/>
             <model type='virtio'/>
           </interface>
           <interface type='network'>
             <source network='default'/>
             <model type='virtio'/>
           </interface>
        </devices>
        """
    )
    ret = virt._diff_interface_lists(old_nics, new_nics)
    assert [nic.find("mac").get("address") for nic in ret["unchanged"]] == [
        "52:54:00:03:02:15",
        "52:54:00:ea:2e:89",
    ]


def test_update_nic_hostdev_nochange(make_mock_network, make_mock_vm, test):
    """
    Test the virt.update function with a running host with hostdev nic
    """
    xml_def_template = """
        <domain type='kvm'>
          <name>my_vm</name>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <os>
            <type arch='x86_64'>hvm</type>
          </os>
          <on_reboot>restart</on_reboot>
          <devices>
            {}
          </devices>
        </domain>
    """
    inactive_nic = """
        <interface type='hostdev' managed='yes'>
          <mac address='52:54:00:67:b2:08'/>
          <driver name='vfio'/>
          <source network="test-hostdev"/>
          <model type='virtio'/>
          <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
        </interface>
    """
    running_nic = """
        <interface type='hostdev' managed='yes'>
          <mac address='52:54:00:67:b2:08'/>
          <driver name='vfio'/>
          <source>
            <address type='pci' domain='0x0000' bus='0x3d' slot='0x02' function='0x0'/>
          </source>
          <model type='virtio'/>
          <alias name='hostdev0'/>
          <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
        </interface>
    """
    domain_mock = make_mock_vm(
        xml_def_template.format(running_nic),
        running="running",
        inactive_def=xml_def_template.format(inactive_nic),
    )

    make_mock_network(
        """
        <network connections='1'>
          <name>test-hostdev</name>
          <uuid>51d0aaa5-7530-4c60-8498-5bc3ab8c655b</uuid>
          <forward mode='hostdev' managed='yes'>
            <pf dev='eth0'/>
            <address type='pci' domain='0x0000' bus='0x3d' slot='0x02' function='0x0'/>
            <address type='pci' domain='0x0000' bus='0x3d' slot='0x02' function='0x1'/>
          </forward>
        </network>
        """
    )

    ret = virt.update(
        "my_vm",
        interfaces=[{"name": "eth0", "type": "network", "source": "test-hostdev"}],
        test=test,
        live=True,
    )
    assert not ret.get("definition")
    assert not ret.get("interface").get("attached")
    assert not ret.get("interface").get("detached")
    define_mock = virt.libvirt.openAuth().defineXML
    define_mock.assert_not_called()
    domain_mock.attachDevice.assert_not_called()
    domain_mock.detachDevice.assert_not_called()


def test_update_no_param(make_mock_vm):
    """
    Test virt.update(), no parameter passed
    """
    domain_mock = make_mock_vm()
    ret = virt.update("my_vm")
    assert not ret["definition"]
    assert not ret.get("mem")
    assert not ret.get("cpu")


def test_update_cpu_and_mem(make_mock_vm):
    """
    Test virt.update(), update both cpu and memory
    """
    domain_mock = make_mock_vm()
    ret = virt.update("my_vm", mem=2048, cpu=2)
    assert ret["definition"]
    assert ret["mem"]
    assert ret["cpu"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("vcpu").text == "2"
    assert setxml.find("memory").text == "2147483648"
    assert domain_mock.setMemoryFlags.call_args[0][0] == 2048 * 1024
    assert domain_mock.setVcpusFlags.call_args[0][0] == 2


def test_update_cpu_simple(make_mock_vm):
    """
    Test virt.update(), simple cpu count update
    """
    domain_mock = make_mock_vm()
    ret = virt.update("my_vm", cpu=2)
    assert ret["definition"]
    assert ret["cpu"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("vcpu").text == "2"
    assert domain_mock.setVcpusFlags.call_args[0][0] == 2


def test_update_autostart(make_mock_vm):
    """
    Test virt.update(), simple autostart update
    """
    domain_mock = make_mock_vm()
    virt.update("my_vm", autostart=True)
    domain_mock.setAutostart.assert_called_with(1)


def test_update_add_cpu_topology(make_mock_vm):
    """
    Test virt.update(), add cpu topology settings
    """
    domain_mock = make_mock_vm()
    ret = virt.update(
        "my_vm",
        cpu={
            "placement": "static",
            "cpuset": "0-11",
            "current": 5,
            "maximum": 12,
            "vcpus": {
                "0": {"enabled": True, "hotpluggable": False, "order": 1},
                "1": {"enabled": False, "hotpluggable": True},
            },
            "mode": "custom",
            "match": "exact",
            "check": "full",
            "model": {
                "name": "coreduo",
                "fallback": "allow",
                "vendor_id": "Genuine20201",
            },
            "vendor": "Intel",
            "topology": {"sockets": 1, "cores": 12, "threads": 1},
            "cache": {"mode": "emulate", "level": 3},
            "features": {"lahf": "optional", "pcid": "disable"},
            "numa": {
                "0": {
                    "cpus": "0-3",
                    "memory": "1g",
                    "discard": True,
                    "distances": {0: 10, 1: 21, 2: 31, 3: 41},
                },
                "1": {
                    "cpus": "4-6",
                    "memory": "0.5g",
                    "discard": False,
                    "memAccess": "shared",
                    "distances": {0: 21, 1: 10, 2: 15, 3: 30},
                },
            },
        },
    )
    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])

    assert setxml.find("vcpu").text == "12"
    assert setxml.find("vcpu").get("placement") == "static"
    assert setxml.find("vcpu").get("cpuset") == "0,1,2,3,4,5,6,7,8,9,10,11"
    assert setxml.find("vcpu").get("current") == "5"

    assert setxml.find("./vcpus/vcpu/[@id='0']").get("id") == "0"
    assert setxml.find("./vcpus/vcpu/[@id='0']").get("enabled") == "yes"
    assert setxml.find("./vcpus/vcpu/[@id='0']").get("hotpluggable") == "no"
    assert setxml.find("./vcpus/vcpu/[@id='0']").get("order") == "1"
    assert setxml.find("./vcpus/vcpu/[@id='1']").get("id") == "1"
    assert setxml.find("./vcpus/vcpu/[@id='1']").get("enabled") == "no"
    assert setxml.find("./vcpus/vcpu/[@id='1']").get("hotpluggable") == "yes"
    assert setxml.find("./vcpus/vcpu/[@id='1']").get("order") is None

    assert setxml.find("cpu").get("mode") == "custom"
    assert setxml.find("cpu").get("match") == "exact"
    assert setxml.find("cpu").get("check") == "full"

    assert setxml.find("cpu/model").get("vendor_id") == "Genuine20201"
    assert setxml.find("cpu/model").get("fallback") == "allow"
    assert setxml.find("cpu/model").text == "coreduo"

    assert setxml.find("cpu/vendor").text == "Intel"

    assert setxml.find("cpu/topology").get("sockets") == "1"
    assert setxml.find("cpu/topology").get("cores") == "12"
    assert setxml.find("cpu/topology").get("threads") == "1"

    assert setxml.find("cpu/cache").get("level") == "3"
    assert setxml.find("cpu/cache").get("mode") == "emulate"

    assert setxml.find("./cpu/feature[@name='pcid']").get("policy") == "disable"
    assert setxml.find("./cpu/feature[@name='lahf']").get("policy") == "optional"

    assert setxml.find("./cpu/numa/cell/[@id='0']").get("cpus") == "0,1,2,3"
    assert setxml.find("./cpu/numa/cell/[@id='0']").get("memory") == str(1024**3)
    assert setxml.find("./cpu/numa/cell/[@id='0']").get("unit") == "bytes"
    assert setxml.find("./cpu/numa/cell/[@id='0']").get("discard") == "yes"
    assert (
        setxml.find("./cpu/numa/cell/[@id='0']/distances/sibling/[@id='0']").get(
            "value"
        )
        == "10"
    )
    assert (
        setxml.find("./cpu/numa/cell/[@id='0']/distances/sibling/[@id='1']").get(
            "value"
        )
        == "21"
    )
    assert (
        setxml.find("./cpu/numa/cell/[@id='0']/distances/sibling/[@id='2']").get(
            "value"
        )
        == "31"
    )
    assert (
        setxml.find("./cpu/numa/cell/[@id='0']/distances/sibling/[@id='3']").get(
            "value"
        )
        == "41"
    )
    assert setxml.find("./cpu/numa/cell/[@id='1']").get("cpus") == "4,5,6"
    assert setxml.find("./cpu/numa/cell/[@id='1']").get("memory") == str(
        int(1024**3 / 2)
    )
    assert setxml.find("./cpu/numa/cell/[@id='1']").get("unit") == "bytes"
    assert setxml.find("./cpu/numa/cell/[@id='1']").get("discard") == "no"
    assert setxml.find("./cpu/numa/cell/[@id='1']").get("memAccess") == "shared"
    assert (
        setxml.find("./cpu/numa/cell/[@id='1']/distances/sibling/[@id='0']").get(
            "value"
        )
        == "21"
    )
    assert (
        setxml.find("./cpu/numa/cell/[@id='1']/distances/sibling/[@id='1']").get(
            "value"
        )
        == "10"
    )
    assert (
        setxml.find("./cpu/numa/cell/[@id='1']/distances/sibling/[@id='2']").get(
            "value"
        )
        == "15"
    )
    assert (
        setxml.find("./cpu/numa/cell/[@id='1']/distances/sibling/[@id='3']").get(
            "value"
        )
        == "30"
    )


@pytest.mark.parametrize("boot_dev", ["hd", "cdrom network hd"])
def test_update_bootdev_unchanged(make_mock_vm, boot_dev):
    """
    Test virt.update(), unchanged boot devices case
    """
    domain_mock = make_mock_vm(
        """
            <domain type='kvm' id='7'>
              <name>my_vm</name>
              <memory unit='KiB'>1048576</memory>
              <currentMemory unit='KiB'>1048576</currentMemory>
              <vcpu placement='auto'>1</vcpu>
              <on_reboot>restart</on_reboot>
              <os>
                <type arch='x86_64' machine='pc-i440fx-2.6'>hvm</type>
                <boot dev="hd"/>
              </os>
            </domain>
        """
    )
    ret = virt.update("my_vm", boot_dev=boot_dev)
    assert ret["definition"] == (boot_dev != "hd")
    if boot_dev == "hd":
        virt.libvirt.openAuth().defineXML.assert_not_called()
    else:
        setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
        assert [node.get("dev") for node in setxml.findall("os/boot")] == [
            "cdrom",
            "network",
            "hd",
        ]


def test_update_boot_kernel_paths(make_mock_vm):
    """
    Test virt.update(), change boot with kernel/initrd path and kernel params
    """
    domain_mock = make_mock_vm()
    ret = virt.update(
        "my_vm",
        boot={
            "kernel": "/root/f8-i386-vmlinuz",
            "initrd": "/root/f8-i386-initrd",
            "cmdline": "console=ttyS0 ks=http://example.com/f8-i386/os/",
        },
    )
    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("os/kernel").text == "/root/f8-i386-vmlinuz"
    assert setxml.find("os/initrd").text == "/root/f8-i386-initrd"
    assert (
        setxml.find("os/cmdline").text
        == "console=ttyS0 ks=http://example.com/f8-i386/os/"
    )


def test_update_boot_uefi_paths(make_mock_vm):
    """
    Test virt.update(), add boot with uefi loader and nvram paths
    """
    domain_mock = make_mock_vm()

    ret = virt.update(
        "my_vm",
        boot={
            "loader": "/usr/share/OVMF/OVMF_CODE.fd",
            "nvram": "/usr/share/OVMF/OVMF_VARS.ms.fd",
        },
    )

    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("os/loader").text == "/usr/share/OVMF/OVMF_CODE.fd"
    assert setxml.find("os/loader").get("readonly") == "yes"
    assert setxml.find("os/loader").get("type") == "pflash"
    assert setxml.find("os/nvram").get("template") == "/usr/share/OVMF/OVMF_VARS.ms.fd"


def test_update_boot_uefi_auto(make_mock_vm):
    """
    Test virt.update(), change boot with efi value (automatic discovery of loader)
    """
    domain_mock = make_mock_vm()

    ret = virt.update("my_vm", boot={"efi": True})
    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("os").get("firmware") == "efi"


def test_update_boot_uefi_auto_nochange(make_mock_vm):
    """
    Test virt.update(), change boot with efi value and no change.
    libvirt converts the efi=True value into a loader and nvram config with path.
    """
    domain_mock = make_mock_vm(
        """
        <domain type='kvm' id='1'>
          <name>my_vm</name>
          <uuid>27434df0-706d-4603-8ad7-5a88d19a3417</uuid>
          <memory unit='KiB'>524288</memory>
          <currentMemory unit='KiB'>524288</currentMemory>
          <vcpu placement='static'>1</vcpu>
          <resource>
            <partition>/machine</partition>
          </resource>
          <os>
            <type arch='x86_64' machine='pc-i440fx-4.2'>hvm</type>
            <loader readonly='yes' type='pflash'>/usr/share/qemu/edk2-x86_64-code.fd</loader>
            <nvram template='/usr/share/qemu/edk2-i386-vars.fd'>/var/lib/libvirt/qemu/nvram/vm01_VARS.fd</nvram>
          </os>
          <on_reboot>restart</on_reboot>
        </domain>
        """
    )

    ret = virt.update("my_vm", boot={"efi": True})
    assert not ret["definition"]
    virt.libvirt.openAuth().defineXML.assert_not_called()


def test_update_boot_invalid(make_mock_vm):
    """
    Test virt.update(), change boot, invalid values
    """
    domain_mock = make_mock_vm()

    with pytest.raises(SaltInvocationError):
        virt.update(
            "my_vm",
            boot={
                "loader": "/usr/share/OVMF/OVMF_CODE.fd",
                "initrd": "/root/f8-i386-initrd",
            },
        )

    with pytest.raises(SaltInvocationError):
        virt.update("my_vm", boot={"efi": "Not a boolean value"})


def test_update_add_memtune(make_mock_vm):
    """
    Test virt.update(), add memory tune config case
    """
    domain_mock = make_mock_vm()
    ret = virt.update(
        "my_vm",
        mem={
            "soft_limit": "0.5g",
            "hard_limit": "1024",
            "swap_hard_limit": "2048m",
            "min_guarantee": "1 g",
        },
    )

    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert_equal_unit(setxml.find("memtune/soft_limit"), int(0.5 * 1024**3), "bytes")
    assert_equal_unit(setxml.find("memtune/hard_limit"), 1024 * 1024**2, "bytes")
    assert_equal_unit(setxml.find("memtune/swap_hard_limit"), 2048 * 1024**2, "bytes")
    assert_equal_unit(setxml.find("memtune/min_guarantee"), 1 * 1024**3, "bytes")


def test_update_add_memtune_invalid_unit(make_mock_vm):
    """
    Test virt.update(), add invalid unit to memory tuning config
    """
    domain_mock = make_mock_vm()

    with pytest.raises(SaltInvocationError):
        virt.update("my_vm", mem={"soft_limit": "2HB"})

    with pytest.raises(SaltInvocationError):
        virt.update("my_vm", mem={"soft_limit": "3.4.MB"})


def test_update_add_numatune(make_mock_vm):
    """
    Test virt.update(), add numatune config case
    """
    domain_mock = make_mock_vm()
    ret = virt.update(
        "my_vm",
        numatune={
            "memory": {"mode": "strict", "nodeset": 1},
            "memnodes": {
                0: {"mode": "strict", "nodeset": 1},
                1: {"mode": "preferred", "nodeset": 2},
            },
        },
    )

    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("numatune/memory").get("mode") == "strict"
    assert setxml.find("numatune/memory").get("nodeset") == "1"
    assert setxml.find("./numatune/memnode/[@cellid='0']").get("mode") == "strict"
    assert setxml.find("./numatune/memnode/[@cellid='0']").get("nodeset") == "1"
    assert setxml.find("./numatune/memnode/[@cellid='1']").get("mode") == "preferred"
    assert setxml.find("./numatune/memnode/[@cellid='1']").get("nodeset") == "2"


def test_update_mem_simple(make_mock_vm):
    """
    Test virt.update(), simple memory amount change
    """
    domain_mock = make_mock_vm()
    ret = virt.update("my_vm", mem=2048)
    assert ret["definition"]
    assert ret["mem"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("memory").text == str(2048 * 1024**2)
    assert setxml.find("memory").get("unit") == "bytes"
    assert domain_mock.setMemoryFlags.call_args[0][0] == 2048 * 1024


def test_update_mem(make_mock_vm):
    """
    Test virt.update(), advanced memory amounts changes
    """
    domain_mock = make_mock_vm()

    ret = virt.update(
        "my_vm",
        mem={"boot": "0.5g", "current": "2g", "max": "1g", "slots": 12},
    )
    assert ret["definition"]
    assert ret["mem"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("memory").get("unit") == "bytes"
    assert setxml.find("memory").text == str(int(0.5 * 1024**3))
    assert setxml.find("maxMemory").text == str(1 * 1024**3)
    assert setxml.find("currentMemory").text == str(2 * 1024**3)


def test_update_add_mem_backing(make_mock_vm):
    """
    Test virt.update(), add memory backing case
    """
    domain_mock = make_mock_vm()
    ret = virt.update(
        "my_vm",
        mem={
            "hugepages": [
                {"nodeset": "1-5,^4", "size": "1g"},
                {"nodeset": "4", "size": "2g"},
            ],
            "nosharepages": True,
            "locked": True,
            "source": "file",
            "access": "shared",
            "allocation": "immediate",
            "discard": True,
        },
    )

    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert {
        p.get("nodeset"): {"size": p.get("size"), "unit": p.get("unit")}
        for p in setxml.findall("memoryBacking/hugepages/page")
    } == {
        "1,2,3,5": {"size": str(1024**3), "unit": "bytes"},
        "4": {"size": str(2 * 1024**3), "unit": "bytes"},
    }
    assert setxml.find("./memoryBacking/nosharepages") is not None
    assert setxml.find("./memoryBacking/nosharepages").text is None
    assert setxml.find("./memoryBacking/nosharepages").keys() == []
    assert setxml.find("./memoryBacking/locked") is not None
    assert setxml.find("./memoryBacking/locked").text is None
    assert setxml.find("./memoryBacking/locked").keys() == []
    assert setxml.find("./memoryBacking/source").attrib["type"] == "file"
    assert setxml.find("./memoryBacking/access").attrib["mode"] == "shared"
    assert setxml.find("./memoryBacking/discard") is not None


def test_update_add_iothreads(make_mock_vm):
    """
    Test virt.update(), add iothreads
    """
    domain_mock = make_mock_vm()
    ret = virt.update("my_vm", cpu={"iothreads": 5})
    assert ret["definition"]
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("iothreads").text == "5"


def test_update_add_cputune(make_mock_vm):
    """
    Test virt.update(), adding CPU tuning parameters
    """
    domain_mock = make_mock_vm()
    cputune = {
        "shares": 2048,
        "period": 122000,
        "quota": -1,
        "global_period": 1000000,
        "global_quota": -3,
        "emulator_period": 1200000,
        "emulator_quota": -10,
        "iothread_period": 133000,
        "iothread_quota": -1,
        "vcpupin": {0: "1-4,^2", 1: "0,1", 2: "2,3", 3: "0,4"},
        "emulatorpin": "1-3",
        "iothreadpin": {1: "5-6", 2: "7-8"},
        "vcpusched": [
            {"scheduler": "fifo", "priority": 1, "vcpus": "0"},
            {"scheduler": "fifo", "priotity": 2, "vcpus": "1"},
            {"scheduler": "idle", "priotity": 3, "vcpus": "2"},
        ],
        "iothreadsched": [{"scheduler": "batch", "iothreads": "7"}],
        "cachetune": {
            "0-3": {
                0: {"level": 3, "type": "both", "size": 3},
                1: {"level": 3, "type": "both", "size": 3},
                "monitor": {1: 3, "0-3": 3},
            },
            "4-5": {"monitor": {4: 3, 5: 2}},
        },
        "memorytune": {"0-2": {0: 60}, "3-4": {0: 50, 1: 70}},
    }
    assert virt.update("my_vm", cpu={"tuning": cputune}) == {
        "definition": True,
        "disk": {"attached": [], "detached": [], "updated": []},
        "interface": {"attached": [], "detached": []},
    }
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("cputune/shares").text == "2048"
    assert setxml.find("cputune/period").text == "122000"
    assert setxml.find("cputune/quota").text == "-1"
    assert setxml.find("cputune/global_period").text == "1000000"
    assert setxml.find("cputune/global_quota").text == "-3"
    assert setxml.find("cputune/emulator_period").text == "1200000"
    assert setxml.find("cputune/emulator_quota").text == "-10"
    assert setxml.find("cputune/iothread_period").text == "133000"
    assert setxml.find("cputune/iothread_quota").text == "-1"
    assert setxml.find("cputune/vcpupin[@vcpu='0']").get("cpuset") == "1,3,4"
    assert setxml.find("cputune/vcpupin[@vcpu='1']").get("cpuset") == "0,1"
    assert setxml.find("cputune/vcpupin[@vcpu='2']").get("cpuset") == "2,3"
    assert setxml.find("cputune/vcpupin[@vcpu='3']").get("cpuset") == "0,4"
    assert setxml.find("cputune/emulatorpin").get("cpuset") == "1,2,3"
    assert setxml.find("cputune/iothreadpin[@iothread='1']").get("cpuset") == "5,6"
    assert setxml.find("cputune/iothreadpin[@iothread='2']").get("cpuset") == "7,8"
    assert setxml.find("cputune/vcpusched[@vcpus='0']").get("priority") == "1"
    assert setxml.find("cputune/vcpusched[@vcpus='0']").get("scheduler") == "fifo"
    assert setxml.find("cputune/iothreadsched").get("iothreads") == "7"
    assert setxml.find("cputune/iothreadsched").get("scheduler") == "batch"
    assert (
        setxml.find("./cputune/cachetune[@vcpus='0,1,2,3']/cache[@id='0']").get("level")
        == "3"
    )
    assert (
        setxml.find("./cputune/cachetune[@vcpus='0,1,2,3']/cache[@id='0']").get("type")
        == "both"
    )
    assert (
        setxml.find("./cputune/cachetune[@vcpus='0,1,2,3']/monitor[@vcpus='1']").get(
            "level"
        )
        == "3"
    )
    assert (
        setxml.find("./cputune/cachetune[@vcpus='4,5']/monitor[@vcpus='4']").get(
            "level"
        )
        == "3"
    )
    assert (
        setxml.find("./cputune/cachetune[@vcpus='4,5']/monitor[@vcpus='5']").get(
            "level"
        )
        == "2"
    )
    assert (
        setxml.find("./cputune/memorytune[@vcpus='0,1,2']/node[@id='0']").get(
            "bandwidth"
        )
        == "60"
    )
    assert (
        setxml.find("./cputune/memorytune[@vcpus='3,4']/node[@id='0']").get("bandwidth")
        == "50"
    )
    assert (
        setxml.find("./cputune/memorytune[@vcpus='3,4']/node[@id='1']").get("bandwidth")
        == "70"
    )


def test_update_graphics(make_mock_vm):
    """
    Test virt.update(), graphics update case
    """
    domain_mock = make_mock_vm(
        """
        <domain type='kvm' id='7'>
          <name>my_vm</name>
          <memory unit='KiB'>1048576</memory>
          <currentMemory unit='KiB'>1048576</currentMemory>
          <vcpu placement='auto'>1</vcpu>
          <on_reboot>restart</on_reboot>
          <os>
            <type arch='x86_64' machine='pc-i440fx-2.6'>hvm</type>
          </os>
          <devices>
            <graphics type='spice' listen='127.0.0.1' autoport='yes'>
              <listen type='address' address='127.0.0.1'/>
            </graphics>
          </devices>
        </domain>
        """
    )
    assert virt.update("my_vm", graphics={"type": "vnc"}) == {
        "definition": True,
        "disk": {"attached": [], "detached": [], "updated": []},
        "interface": {"attached": [], "detached": []},
    }
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("devices/graphics").get("type") == "vnc"


def test_update_console(make_mock_vm):
    """
    Test virt.update(), console and serial devices update case
    """
    domain_mock = make_mock_vm(
        """
        <domain type='kvm' id='7'>
          <name>my_vm</name>
          <memory unit='KiB'>1048576</memory>
          <currentMemory unit='KiB'>1048576</currentMemory>
          <vcpu placement='auto'>1</vcpu>
          <on_reboot>restart</on_reboot>
          <os>
            <type arch='x86_64' machine='pc-i440fx-2.6'>hvm</type>
          </os>
          <devices>
            <serial type='pty'>
              <source path='/dev/pts/4'/>
            </serial>
            <console type='pty'/>
          </devices>
        </domain>
        """
    )

    assert virt.update(
        "my_vm", serials=[{"type": "tcp"}], consoles=[{"type": "tcp"}]
    ) == {
        "definition": True,
        "disk": {"attached": [], "detached": [], "updated": []},
        "interface": {"attached": [], "detached": []},
    }
    setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
    assert setxml.find("devices/serial").attrib["type"] == "tcp"
    assert setxml.find("devices/console").attrib["type"] == "tcp"


def test_update_disks(make_mock_vm):
    """
    Test virt.udpate() with disk device changes
    """
    root_dir = os.path.join(salt.syspaths.ROOT_DIR, "srv", "salt-images")
    xml_def = """
        <domain type='kvm' id='7'>
          <name>my_vm</name>
          <memory unit='KiB'>1048576</memory>
          <currentMemory unit='KiB'>1048576</currentMemory>
          <vcpu placement='auto'>1</vcpu>
          <on_reboot>restart</on_reboot>
          <os>
            <type arch='x86_64' machine='pc-i440fx-2.6'>hvm</type>
          </os>
          <devices>
            <disk type='file' device='disk'>
              <driver name='qemu' type='qcow2'/>
              <source file='{}{}my_vm_system.qcow2'/>
              <backingStore/>
              <target dev='vda' bus='virtio'/>
              <alias name='virtio-disk0'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x07' function='0x0'/>
            </disk>
            <disk type="network" device="disk">
              <driver name='raw' type='qcow2'/>
              <source protocol='rbd' name='libvirt-pool/my_vm_data2'>
                <host name='ses2.tf.local'/>
              </source>
              <target dev='vdc' bus='virtio'/>
              <alias name='virtio-disk2'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x07' function='0x2'/>
            </disk>
          </devices>
        </domain>
    """.format(
        root_dir, os.sep
    )
    domain_mock = make_mock_vm(xml_def)

    mock_chmod = MagicMock()
    mock_run = MagicMock()
    with patch.dict(os.__dict__, {"chmod": mock_chmod, "makedirs": MagicMock()}):
        with patch.dict(virt.__salt__, {"cmd.run": mock_run}):
            ret = virt.update(
                "my_vm",
                disk_profile="default",
                disks=[
                    {
                        "name": "cddrive",
                        "device": "cdrom",
                        "source_file": None,
                        "model": "ide",
                    },
                    {"name": "added", "size": 2048, "io": "threads"},
                ],
            )
            added_disk_path = os.path.join(
                virt.__salt__["config.get"]("virt:images"), "my_vm_added.qcow2"
            )
            assert (
                mock_run.call_args[0][0]
                == f'qemu-img create -f qcow2 "{added_disk_path}" 2048M'
            )
            assert mock_chmod.call_args[0][0] == added_disk_path
            assert [
                (
                    ET.fromstring(disk).find("source").get("file")
                    if str(disk).find("<source") > -1
                    else None
                )
                for disk in ret["disk"]["attached"]
            ] == [None, os.path.join(root_dir, "my_vm_added.qcow2")]

            assert [
                ET.fromstring(disk).find("source").get("volume")
                or ET.fromstring(disk).find("source").get("name")
                for disk in ret["disk"]["detached"]
            ] == ["libvirt-pool/my_vm_data2"]
            assert domain_mock.attachDevice.call_count == 2
            assert domain_mock.detachDevice.call_count == 1

            setxml = ET.fromstring(virt.libvirt.openAuth().defineXML.call_args[0][0])
            assert setxml.find("devices/disk[3]/driver").get("io") == "threads"


def test_update_disks_existing_block(make_mock_vm):
    """
    Test virt.udpate() when adding existing block devices
    """
    root_dir = os.path.join(salt.syspaths.ROOT_DIR, "srv", "salt-images")
    xml_def = """
        <domain type='kvm' id='7'>
          <name>my_vm</name>
          <memory unit='KiB'>1048576</memory>
          <currentMemory unit='KiB'>1048576</currentMemory>
          <vcpu placement='auto'>1</vcpu>
          <on_reboot>restart</on_reboot>
          <os>
            <type arch='x86_64' machine='pc-i440fx-2.6'>hvm</type>
          </os>
          <devices>
            <disk type='file' device='disk'>
              <driver name='qemu' type='qcow2'/>
              <source file='{}{}my_vm_system.qcow2'/>
              <backingStore/>
              <target dev='vda' bus='virtio'/>
              <alias name='virtio-disk0'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x07' function='0x0'/>
            </disk>
          </devices>
        </domain>
    """.format(
        root_dir, os.sep
    )
    domain_mock = make_mock_vm(xml_def)

    mock_chmod = MagicMock()
    mock_run = MagicMock()
    with patch.dict(os.__dict__, {"chmod": mock_chmod, "makedirs": MagicMock()}):
        with patch.dict(
            os.path.__dict__,
            {
                "exists": MagicMock(return_value=True),
                "isfile": MagicMock(return_value=False),
            },
        ):
            with patch.dict(virt.__salt__, {"cmd.run": mock_run}):
                ret = virt.update(
                    "my_vm",
                    disk_profile="default",
                    disks=[
                        {
                            "name": "data",
                            "format": "raw",
                            "source_file": "/dev/ssd/data",
                        },
                    ],
                )
                assert [
                    (
                        ET.fromstring(disk).find("source").get("file")
                        if str(disk).find("<source") > -1
                        else None
                    )
                    for disk in ret["disk"]["attached"]
                ] == ["/dev/ssd/data"]

                assert domain_mock.attachDevice.call_count == 1
                assert domain_mock.detachDevice.call_count == 0


def test_update_nics(make_mock_vm):
    """
    Test virt.update() with NIC device changes
    """
    domain_mock = make_mock_vm(
        """
        <domain type='kvm' id='7'>
          <name>my_vm</name>
          <memory unit='KiB'>1048576</memory>
          <currentMemory unit='KiB'>1048576</currentMemory>
          <vcpu placement='auto'>1</vcpu>
          <on_reboot>restart</on_reboot>
          <os>
            <type arch='x86_64' machine='pc-i440fx-2.6'>hvm</type>
          </os>
          <devices>
            <interface type='network'>
              <mac address='52:54:00:39:02:b1'/>
              <source network='default' bridge='virbr0'/>
              <target dev='vnet0'/>
              <model type='virtio'/>
              <alias name='net0'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
            </interface>
            <interface type='network'>
              <mac address='52:54:00:39:02:b2'/>
              <source network='oldnet' bridge='virbr1'/>
              <target dev='vnet1'/>
              <model type='virtio'/>
              <alias name='net1'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x1'/>
            </interface>
          </devices>
        </domain>
        """
    )
    mock_config = salt.utils.yaml.safe_load(
        """
          virt:
             nic:
                myprofile:
                   - network: default
                     name: eth0
        """
    )
    with patch.dict(salt.modules.config.__opts__, mock_config):
        ret = virt.update(
            "my_vm",
            nic_profile="myprofile",
            interfaces=[
                {
                    "name": "eth0",
                    "type": "network",
                    "source": "default",
                    "mac": "52:54:00:39:02:b1",
                },
                {"name": "eth1", "type": "network", "source": "newnet"},
            ],
        )
        assert [
            ET.fromstring(nic).find("source").get("network")
            for nic in ret["interface"]["attached"]
        ] == ["newnet"]
        assert [
            ET.fromstring(nic).find("source").get("network")
            for nic in ret["interface"]["detached"]
        ] == ["oldnet"]
        domain_mock.attachDevice.assert_called_once()
        domain_mock.detachDevice.assert_called_once()


def test_update_remove_disks_nics(make_mock_vm):
    """
    Test virt.update() when removing nics and disks even if that may sound silly
    """
    root_dir = os.path.join(salt.syspaths.ROOT_DIR, "srv", "salt-images")
    xml_def = """
        <domain type='kvm' id='7'>
          <name>my_vm</name>
          <memory unit='KiB'>1048576</memory>
          <currentMemory unit='KiB'>1048576</currentMemory>
          <vcpu placement='auto'>1</vcpu>
          <on_reboot>restart</on_reboot>
          <os>
            <type arch='x86_64' machine='pc-i440fx-2.6'>hvm</type>
          </os>
          <devices>
            <disk type='file' device='disk'>
              <driver name='qemu' type='qcow2'/>
              <source file='{}{}my_vm_system.qcow2'/>
              <backingStore/>
              <target dev='vda' bus='virtio'/>
              <alias name='virtio-disk0'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x07' function='0x0'/>
            </disk>
            <interface type='network'>
              <mac address='52:54:00:39:02:b1'/>
              <source network='default' bridge='virbr0'/>
              <target dev='vnet0'/>
              <model type='virtio'/>
              <alias name='net0'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
            </interface>
          </devices>
        </domain>
    """.format(
        root_dir, os.sep
    )
    domain_mock = make_mock_vm(xml_def)

    ret = virt.update(
        "my_vm", nic_profile=None, interfaces=[], disk_profile=None, disks=[]
    )
    assert ret["interface"].get("attached", []) == []
    assert len(ret["interface"]["detached"]) == 1
    assert ret["disk"].get("attached", []) == []
    assert len(ret["disk"]["detached"]) == 1

    domain_mock.attachDevice.assert_not_called()
    assert domain_mock.detachDevice.call_count == 2


def test_update_no_change(make_mock_vm, make_mock_storage_pool):
    """
    Test virt.update() with no change
    """
    root_dir = os.path.join(salt.syspaths.ROOT_DIR, "srv", "salt-images")
    xml_def = """
        <domain type='kvm' id='7'>
          <name>my_vm</name>
          <memory unit='KiB'>1048576</memory>
          <currentMemory unit='KiB'>1048576</currentMemory>
          <vcpu placement='auto'>1</vcpu>
          <on_reboot>restart</on_reboot>
          <os>
            <type arch='x86_64' machine='pc-i440fx-2.6'>hvm</type>
            <boot dev="hd"/>
          </os>
          <devices>
            <disk type='file' device='disk'>
              <driver name='qemu' type='qcow2'/>
              <source file='{}{}my_vm_system.qcow2'/>
              <backingStore/>
              <target dev='vda' bus='virtio'/>
              <alias name='virtio-disk0'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x07' function='0x0'/>
            </disk>
            <disk type='volume' device='disk'>
              <driver name='qemu' type='qcow2'/>
              <source pool='default' volume='my_vm_data'/>
              <backingStore/>
              <target dev='vdb' bus='virtio'/>
              <alias name='virtio-disk1'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x07' function='0x1'/>
            </disk>
            <disk type="network" device="disk">
              <driver name='raw' type='qcow2'/>
              <source protocol='rbd' name='libvirt-pool/my_vm_data2'>
                <host name='ses2.tf.local'/>
                <host name='ses3.tf.local' port='1234'/>
                <auth username='libvirt'>
                  <secret type='ceph' usage='pool_test-rbd'/>
                </auth>
              </source>
              <target dev='vdc' bus='virtio'/>
              <alias name='virtio-disk2'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x07' function='0x2'/>
            </disk>
            <interface type='network'>
              <mac address='52:54:00:39:02:b1'/>
              <source network='default' bridge='virbr0'/>
              <target dev='vnet0'/>
              <model type='virtio'/>
              <alias name='net0'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
            </interface>
            <graphics type='spice' listen='127.0.0.1' autoport='yes'>
              <listen type='address' address='127.0.0.1'/>
            </graphics>
            <video>
              <model type='qxl' ram='65536' vram='65536' vgamem='16384' heads='1' primary='yes'/>
              <alias name='video0'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x02' function='0x0'/>
            </video>
            <serial type='pty'/>
            <console type='pty'/>
          </devices>
        </domain>
    """.format(
        root_dir, os.sep
    )
    domain_mock = make_mock_vm(xml_def)

    make_mock_storage_pool("default", "dir", ["my_vm_data"])
    make_mock_storage_pool(
        "test-rbd",
        "rbd",
        ["my_vm_data2"],
        source="""
            <host name='ses2.tf.local'/>
            <host name='ses3.tf.local' port='1234'/>
            <name>libvirt-pool</name>
            <auth type='ceph' username='libvirt'>
              <secret usage='pool_test-rbd'/>
            </auth>
        """,
    )
    assert virt.update(
        "my_vm",
        cpu=1,
        mem=1024,
        disk_profile="default",
        disks=[
            {"name": "data", "size": 2048, "pool": "default"},
            {"name": "data2", "size": 4096, "pool": "test-rbd", "format": "raw"},
        ],
        nic_profile="myprofile",
        interfaces=[
            {
                "name": "eth0",
                "type": "network",
                "source": "default",
                "mac": "52:54:00:39:02:b1",
            },
        ],
        graphics={
            "type": "spice",
            "listen": {"type": "address", "address": "127.0.0.1"},
        },
    ) == {
        "definition": False,
        "disk": {"attached": [], "detached": [], "updated": []},
        "interface": {"attached": [], "detached": []},
    }


def test_update_failure(make_mock_vm):
    """
    Test virt.update() with errors
    """
    domain_mock = make_mock_vm()
    virt.libvirt.openAuth().defineXML.side_effect = virt.libvirt.libvirtError(
        "Test error"
    )
    with pytest.raises(virt.libvirt.libvirtError):
        virt.update("my_vm", mem=2048)

    # Failed single update failure case
    virt.libvirt.openAuth().defineXML = MagicMock(return_value=True)
    domain_mock.setMemoryFlags.side_effect = virt.libvirt.libvirtError(
        "Failed to live change memory"
    )

    domain_mock.setVcpusFlags.return_value = 0
    assert virt.update("my_vm", cpu=4, mem=2048) == {
        "definition": True,
        "errors": ["Failed to live change memory"],
        "cpu": True,
        "disk": {"attached": [], "detached": [], "updated": []},
        "interface": {"attached": [], "detached": []},
    }


@pytest.mark.parametrize("hypervisor", ["kvm", "xen"])
def test_gen_xml_spice_default(hypervisor):
    """
    Test virt._gen_xml() with default spice graphics device
    """
    xml_data = virt._gen_xml(
        virt.libvirt.openAuth.return_value,
        "hello",
        1,
        512,
        {},
        {},
        hypervisor,
        "hvm",
        "x86_64",
        graphics={"type": "spice"},
    )
    root = ET.fromstring(xml_data)
    assert root.find("devices/graphics").attrib["type"] == "spice"
    assert root.find("devices/graphics").attrib["autoport"] == "yes"
    assert root.find("devices/graphics").attrib["listen"] == "0.0.0.0"
    assert root.find("devices/graphics/listen").attrib["type"] == "address"
    assert root.find("devices/graphics/listen").attrib["address"] == "0.0.0.0"
    if hypervisor == "kvm":
        assert (
            root.find(".//channel[@type='spicevmc']/target").get("name")
            == "com.redhat.spice.0"
        )
    else:
        assert root.find(".//channel[@type='spicevmc']") is None


def test_gen_xml_spice():
    """
    Test virt._gen_xml() with spice graphics device
    """
    xml_data = virt._gen_xml(
        virt.libvirt.openAuth.return_value,
        "hello",
        1,
        512,
        {},
        {},
        "kvm",
        "hvm",
        "x86_64",
        graphics={
            "type": "spice",
            "port": 1234,
            "tls_port": 5678,
            "listen": {"type": "none"},
        },
    )
    root = ET.fromstring(xml_data)
    assert root.find("devices/graphics").attrib["type"] == "spice"
    assert root.find("devices/graphics").attrib["autoport"] == "no"
    assert root.find("devices/graphics").attrib["port"] == "1234"
    assert root.find("devices/graphics").attrib["tlsPort"] == "5678"
    assert "listen" not in root.find("devices/graphics").attrib
    assert root.find("devices/graphics/listen").attrib["type"] == "none"
    assert "address" not in root.find("devices/graphics/listen").attrib



import pytest

import salt.states.win_appx as win_appx
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {win_appx: {}}


def test_absent_missing():
    expected = {
        "comment": "No apps found matching query: *candy*",
        "changes": {},
        "name": "remove_candy",
        "result": True,
    }
    mock_list = MagicMock(return_value=["package1", "package2"])
    with patch.dict(win_appx.__salt__, {"appx.list": mock_list}):
        result = win_appx.absent("remove_candy", "*candy*")
        assert result == expected


def test_absent_test_true():
    expected = {
        "comment": "The following apps will be removed:\n- king.com.CandyCrush",
        "changes": {},
        "name": "remove_candy",
        "result": None,
    }
    mock_list = MagicMock(return_value=["package1", "king.com.CandyCrush"])
    with patch.dict(win_appx.__salt__, {"appx.list": mock_list}):
        with patch.dict(win_appx.__opts__, {"test": True}):
            result = win_appx.absent("remove_candy", "*candy*")
            assert result == expected


def test_absent_missing_after_test():
    expected = {
        "comment": "No apps found matching query: *candy*",
        "changes": {},
        "name": "remove_candy",
        "result": False,
    }
    mock_list = MagicMock(return_value=["package1", "king.com.CandyCrush"])
    with patch.dict(
        win_appx.__salt__,
        {
            "appx.list": mock_list,
            "appx.remove": MagicMock(return_value=None),
        },
    ):
        with patch.dict(win_appx.__opts__, {"test": False}):
            result = win_appx.absent("remove_candy", "*candy*")
            assert result == expected


def test_absent():
    expected = {
        "comment": "Removed apps matching query: *candy*",
        "changes": {"old": ["king.com.CandyCrush"]},
        "name": "remove_candy",
        "result": True,
    }
    mock_list = MagicMock(
        side_effect=(
            ["package1", "king.com.CandyCrush"],
            ["package1"],
        ),
    )
    with patch.dict(
        win_appx.__salt__,
        {
            "appx.list": mock_list,
            "appx.remove": MagicMock(return_value=True),
        },
    ):
        with patch.dict(win_appx.__opts__, {"test": False}):
            result = win_appx.absent("remove_candy", "*candy*")
            assert result == expected



"""
Tests for the Mandrill execution module.
"""

import pytest

import salt.modules.mandrill as mandrill
from tests.support.mock import MagicMock, patch


@pytest.fixture
def fn_test_send():
    return {
        "result": True,
        "comment": "",
        "out": [
            {
                "status": "sent",
                "_id": "c4353540a3c123eca112bbdd704ab6",
                "email": "recv@example.com",
                "reject_reason": None,
            }
        ],
    }


@pytest.fixture
def configure_loader_modules():
    module_globals = {
        mandrill: {
            "__salt__": {
                "config.merge": MagicMock(
                    return_value={"mandrill": {"key": "2orgk34kgk34g"}}
                )
            }
        }
    }
    if mandrill.HAS_REQUESTS is False:
        module_globals["sys.modules"] = {"requests": MagicMock()}
    return module_globals


def test_send(fn_test_send):
    """
    Test the send function.
    """
    mock_cmd = MagicMock(return_value=fn_test_send)
    with patch.object(mandrill, "send", mock_cmd) as send:
        assert (
            send(
                message={
                    "subject": "Hi",
                    "from_email": "test@example.com",
                    "to": [{"email": "recv@example.com", "type": "to"}],
                }
            )
            == fn_test_send
        )


def test_deprecation_58640():
    # check that type error will be raised
    message = {
        "subject": "Hi",
        "from_email": "test@example.com",
        "to": [{"email": "recv@example.com", "type": "to"}],
    }
    pytest.raises(TypeError, mandrill.send, **{"message": message, "async": True})

    # check that async will raise an error
    try:
        mandrill.send(  # pylint: disable=unexpected-keyword-arg
            **{"message": message, "async": True}
        )
    except TypeError as no_async:
        assert str(no_async) == "send() got an unexpected keyword argument 'async'"



"""
Control Modjk via the Apache Tomcat "Status" worker
(http://tomcat.apache.org/connectors-doc/reference/status.html)

Below is an example of the configuration needed for this module. This
configuration data can be placed either in :ref:`grains
<targeting-grains>` or :ref:`pillar <salt-pillars>`.

If using grains, this can be accomplished :ref:`statically
<static-custom-grains>` or via a :ref:`grain module <writing-grains>`.

If using pillar, the yaml configuration can be placed directly into a pillar
SLS file, making this both the easier and more dynamic method of configuring
this module.

.. code-block:: yaml

    modjk:
      default:
        url: http://localhost/jkstatus
        user: modjk
        pass: secret
        realm: authentication realm for digest passwords
        timeout: 5
      otherVhost:
        url: http://otherVhost/jkstatus
        user: modjk
        pass: secret2
        realm: authentication realm2 for digest passwords
        timeout: 600
"""

import urllib.parse
import urllib.request


def __virtual__():
    """
    Always load
    """
    return True


def _auth(url, user, passwd, realm):
    """
    returns a authentication handler.
    """

    basic = urllib.request.HTTPBasicAuthHandler()
    basic.add_password(realm=realm, uri=url, user=user, passwd=passwd)
    digest = urllib.request.HTTPDigestAuthHandler()
    digest.add_password(realm=realm, uri=url, user=user, passwd=passwd)
    return urllib.request.build_opener(basic, digest)


def _do_http(opts, profile="default"):
    """
    Make the http request and return the data
    """

    ret = {}

    url = __salt__["config.get"](f"modjk:{profile}:url", "")
    user = __salt__["config.get"](f"modjk:{profile}:user", "")
    passwd = __salt__["config.get"](f"modjk:{profile}:pass", "")
    realm = __salt__["config.get"](f"modjk:{profile}:realm", "")
    timeout = __salt__["config.get"](f"modjk:{profile}:timeout", "")

    if not url:
        raise Exception(f"missing url in profile {profile}")

    if user and passwd:
        auth = _auth(url=url, realm=realm, user=user, passwd=passwd)
        urllib.request.install_opener(auth)

    url += f"?{urllib.parse.urlencode(opts)}"

    for line in urllib.request.urlopen(url, timeout=timeout).read().splitlines():
        splt = line.split("=", 1)
        if splt[0] in ret:
            ret[splt[0]] += f",{splt[1]}"
        else:
            ret[splt[0]] = splt[1]

    return ret


def _worker_ctl(worker, lbn, vwa, profile="default"):
    """
    enable/disable/stop a worker
    """

    cmd = {
        "cmd": "update",
        "mime": "prop",
        "w": lbn,
        "sw": worker,
        "vwa": vwa,
    }
    return _do_http(cmd, profile)["worker.result.type"] == "OK"


def version(profile="default"):
    """
    Return the modjk version

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.version
        salt '*' modjk.version other-profile
    """

    cmd = {
        "cmd": "version",
        "mime": "prop",
    }
    return _do_http(cmd, profile)["worker.jk_version"].split("/")[-1]


def get_running(profile="default"):
    """
    Get the current running config (not from disk)

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.get_running
        salt '*' modjk.get_running other-profile
    """

    cmd = {
        "cmd": "list",
        "mime": "prop",
    }
    return _do_http(cmd, profile)


def dump_config(profile="default"):
    """
    Dump the original configuration that was loaded from disk

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.dump_config
        salt '*' modjk.dump_config other-profile
    """

    cmd = {
        "cmd": "dump",
        "mime": "prop",
    }
    return _do_http(cmd, profile)


def list_configured_members(lbn, profile="default"):
    """
    Return a list of member workers from the configuration files

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.list_configured_members loadbalancer1
        salt '*' modjk.list_configured_members loadbalancer1 other-profile
    """

    config = dump_config(profile)

    try:
        ret = config[f"worker.{lbn}.balance_workers"]
    except KeyError:
        return []

    return [_f for _f in ret.strip().split(",") if _f]


def workers(profile="default"):
    """
    Return a list of member workers and their status

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.workers
        salt '*' modjk.workers other-profile
    """

    config = get_running(profile)
    lbn = config["worker.list"].split(",")
    worker_list = []
    ret = {}

    for lb in lbn:
        try:
            worker_list.extend(config[f"worker.{lb}.balance_workers"].split(","))
        except KeyError:
            pass

    worker_list = list(set(worker_list))

    for worker in worker_list:
        ret[worker] = {
            "activation": config[f"worker.{worker}.activation"],
            "state": config[f"worker.{worker}.state"],
        }

    return ret


def recover_all(lbn, profile="default"):
    """
    Set the all the workers in lbn to recover and activate them if they are not

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.recover_all loadbalancer1
        salt '*' modjk.recover_all loadbalancer1 other-profile
    """

    ret = {}
    config = get_running(profile)
    try:
        workers_ = config[f"worker.{lbn}.balance_workers"].split(",")
    except KeyError:
        return ret

    for worker in workers_:
        curr_state = worker_status(worker, profile)
        if curr_state["activation"] != "ACT":
            worker_activate(worker, lbn, profile)
        if not curr_state["state"].startswith("OK"):
            worker_recover(worker, lbn, profile)
        ret[worker] = worker_status(worker, profile)

    return ret


def reset_stats(lbn, profile="default"):
    """
    Reset all runtime statistics for the load balancer

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.reset_stats loadbalancer1
        salt '*' modjk.reset_stats loadbalancer1 other-profile
    """

    cmd = {
        "cmd": "reset",
        "mime": "prop",
        "w": lbn,
    }
    return _do_http(cmd, profile)["worker.result.type"] == "OK"


def lb_edit(lbn, settings, profile="default"):
    """
    Edit the loadbalancer settings

    Note: http://tomcat.apache.org/connectors-doc/reference/status.html
    Data Parameters for the standard Update Action

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.lb_edit loadbalancer1 "{'vlr': 1, 'vlt': 60}"
        salt '*' modjk.lb_edit loadbalancer1 "{'vlr': 1, 'vlt': 60}" other-profile
    """

    settings["cmd"] = "update"
    settings["mime"] = "prop"
    settings["w"] = lbn

    return _do_http(settings, profile)["worker.result.type"] == "OK"


def bulk_stop(workers, lbn, profile="default"):
    """
    Stop all the given workers in the specific load balancer

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.bulk_stop node1,node2,node3 loadbalancer1
        salt '*' modjk.bulk_stop node1,node2,node3 loadbalancer1 other-profile

        salt '*' modjk.bulk_stop ["node1","node2","node3"] loadbalancer1
        salt '*' modjk.bulk_stop ["node1","node2","node3"] loadbalancer1 other-profile
    """

    ret = {}

    if isinstance(workers, str):
        workers = workers.split(",")

    for worker in workers:
        try:
            ret[worker] = worker_stop(worker, lbn, profile)
        except Exception:  # pylint: disable=broad-except
            ret[worker] = False

    return ret


def bulk_activate(workers, lbn, profile="default"):
    """
    Activate all the given workers in the specific load balancer

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.bulk_activate node1,node2,node3 loadbalancer1
        salt '*' modjk.bulk_activate node1,node2,node3 loadbalancer1 other-profile

        salt '*' modjk.bulk_activate ["node1","node2","node3"] loadbalancer1
        salt '*' modjk.bulk_activate ["node1","node2","node3"] loadbalancer1 other-profile
    """

    ret = {}

    if isinstance(workers, str):
        workers = workers.split(",")

    for worker in workers:
        try:
            ret[worker] = worker_activate(worker, lbn, profile)
        except Exception:  # pylint: disable=broad-except
            ret[worker] = False

    return ret


def bulk_disable(workers, lbn, profile="default"):
    """
    Disable all the given workers in the specific load balancer

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.bulk_disable node1,node2,node3 loadbalancer1
        salt '*' modjk.bulk_disable node1,node2,node3 loadbalancer1 other-profile

        salt '*' modjk.bulk_disable ["node1","node2","node3"] loadbalancer1
        salt '*' modjk.bulk_disable ["node1","node2","node3"] loadbalancer1 other-profile
    """

    ret = {}

    if isinstance(workers, str):
        workers = workers.split(",")

    for worker in workers:
        try:
            ret[worker] = worker_disable(worker, lbn, profile)
        except Exception:  # pylint: disable=broad-except
            ret[worker] = False

    return ret


def bulk_recover(workers, lbn, profile="default"):
    """
    Recover all the given workers in the specific load balancer

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.bulk_recover node1,node2,node3 loadbalancer1
        salt '*' modjk.bulk_recover node1,node2,node3 loadbalancer1 other-profile

        salt '*' modjk.bulk_recover ["node1","node2","node3"] loadbalancer1
        salt '*' modjk.bulk_recover ["node1","node2","node3"] loadbalancer1 other-profile
    """

    ret = {}

    if isinstance(workers, str):
        workers = workers.split(",")

    for worker in workers:
        try:
            ret[worker] = worker_recover(worker, lbn, profile)
        except Exception:  # pylint: disable=broad-except
            ret[worker] = False

    return ret


def worker_status(worker, profile="default"):
    """
    Return the state of the worker

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.worker_status node1
        salt '*' modjk.worker_status node1 other-profile
    """

    config = get_running(profile)
    try:
        return {
            "activation": config[f"worker.{worker}.activation"],
            "state": config[f"worker.{worker}.state"],
        }
    except KeyError:
        return False


def worker_recover(worker, lbn, profile="default"):
    """
    Set the worker to recover
    this module will fail if it is in OK state

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.worker_recover node1 loadbalancer1
        salt '*' modjk.worker_recover node1 loadbalancer1 other-profile
    """

    cmd = {
        "cmd": "recover",
        "mime": "prop",
        "w": lbn,
        "sw": worker,
    }
    return _do_http(cmd, profile)


def worker_disable(worker, lbn, profile="default"):
    """
    Set the worker to disable state in the lbn load balancer

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.worker_disable node1 loadbalancer1
        salt '*' modjk.worker_disable node1 loadbalancer1 other-profile
    """

    return _worker_ctl(worker, lbn, "d", profile)


def worker_activate(worker, lbn, profile="default"):
    """
    Set the worker to activate state in the lbn load balancer

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.worker_activate node1 loadbalancer1
        salt '*' modjk.worker_activate node1 loadbalancer1 other-profile
    """

    return _worker_ctl(worker, lbn, "a", profile)


def worker_stop(worker, lbn, profile="default"):
    """
    Set the worker to stopped state in the lbn load balancer

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.worker_activate node1 loadbalancer1
        salt '*' modjk.worker_activate node1 loadbalancer1 other-profile
    """

    return _worker_ctl(worker, lbn, "s", profile)


def worker_edit(worker, lbn, settings, profile="default"):
    """
    Edit the worker settings

    Note: http://tomcat.apache.org/connectors-doc/reference/status.html
    Data Parameters for the standard Update Action

    CLI Examples:

    .. code-block:: bash

        salt '*' modjk.worker_edit node1 loadbalancer1 "{'vwf': 500, 'vwd': 60}"
        salt '*' modjk.worker_edit node1 loadbalancer1 "{'vwf': 500, 'vwd': 60}" other-profile
    """

    settings["cmd"] = "update"
    settings["mime"] = "prop"
    settings["w"] = lbn
    settings["sw"] = worker

    return _do_http(settings, profile)["worker.result.type"] == "OK"



"""
Validate the cache package functions.
"""

import pytest

import salt.cache
import salt.payload
from tests.support.mock import patch


@pytest.fixture
def opts():
    return {
        "cache": "localfs",
        "memcache_expire_seconds": 0,
        "memcache_max_items": 0,
        "memcache_full_cleanup": False,
        "memcache_debug": False,
    }


def test_factory_cache(opts):
    ret = salt.cache.factory(opts)
    assert isinstance(ret, salt.cache.Cache)


def test_factory_memcache(opts):
    with patch.dict(opts, {"memcache_expire_seconds": 10}):
        ret = salt.cache.factory(opts)
        assert isinstance(ret, salt.cache.MemCache)



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import textwrap

import pytest

import salt.modules.ddns as ddns
import salt.utils.json
from tests.support.mock import MagicMock, mock_open, patch

try:
    import dns.query
    import dns.tsigkeyring

    HAS_DNS = True
except ImportError:
    HAS_DNS = False


pytestmark = [
    pytest.mark.skipif(
        HAS_DNS is False,
        reason="dnspython libs not installed.",
    )
]


@pytest.fixture
def configure_loader_modules():
    return {ddns: {}}


def test_add_host():
    """
    Test cases for Add, replace, or update the A
    and PTR (reverse) records for a host.
    """
    with patch("salt.modules.ddns.update") as ddns_update:
        ddns_update.return_value = False
        assert not ddns.add_host(zone="A", name="B", ttl=1, ip="172.27.0.0")

        ddns_update.return_value = True
        assert ddns.add_host(zone="A", name="B", ttl=1, ip="172.27.0.0")


def test_delete_host():
    """
    Tests for delete the forward and reverse records for a host.
    """
    with patch("salt.modules.ddns.delete") as ddns_delete:
        ddns_delete.return_value = False
        with patch.object(dns.query, "udp") as mock:
            mock.answer = [{"address": "localhost"}]
            assert not ddns.delete_host(zone="A", name="B")


def test_update():
    """
    Test to add, replace, or update a DNS record.
    """
    mock_request = textwrap.dedent(
        """\
        id 29380
        opcode QUERY
        rcode NOERROR
        flags RD
        ;QUESTION
        name.zone. IN AAAA
        ;ANSWER
        ;AUTHORITY
        ;ADDITIONAL"""
    )
    mock_rdtype = 28  # rdtype of AAAA record

    class MockRrset:
        def __init__(self):
            self.items = [{"address": "localhost"}]
            self.ttl = 2

    class MockAnswer:
        def __init__(self, *args, **kwargs):
            self.answer = [MockRrset()]

        def rcode(self):
            return 0

    def mock_udp_query(*args, **kwargs):
        return MockAnswer

    with patch.object(dns.message, "make_query", MagicMock(return_value=mock_request)):
        with patch.object(dns.query, "udp", mock_udp_query()):
            with patch.object(
                dns.rdatatype, "from_text", MagicMock(return_value=mock_rdtype)
            ):
                with patch.object(ddns, "_get_keyring", return_value=None):
                    with patch.object(ddns, "_config", return_value=None):
                        assert ddns.update("zone", "name", 1, "AAAA", "::1")


def test_delete():
    """
    Test to delete a DNS record.
    """
    file_data = salt.utils.json.dumps({"A": "B"})

    class MockAnswer:
        def __init__(self, *args, **kwargs):
            self.answer = [{"address": "localhost"}]

        def rcode(self):
            return 0

    def mock_udp_query(*args, **kwargs):
        return MockAnswer

    with patch.object(dns.query, "udp", mock_udp_query()):
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=file_data), create=True
        ):
            with patch.object(dns.tsigkeyring, "from_text", return_value=True):
                with patch.object(ddns, "_get_keyring", return_value=None):
                    with patch.object(ddns, "_config", return_value=None):
                        assert ddns.delete(zone="A", name="B")



"""
Management of MySQL databases (schemas)
=======================================

:depends:   - MySQLdb Python module
:configuration: See :py:mod:`salt.modules.mysql` for setup instructions.

The mysql_database module is used to create and manage MySQL databases.
Databases can be set as either absent or present.

.. code-block:: yaml

    frank:
      mysql_database.present
"""

import logging
import sys

log = logging.getLogger(__name__)

# pylint: disable=undefined-variable


def __virtual__():
    """
    Only load if the mysql module is available in __salt__
    """
    if "mysql.db_exists" in __salt__:
        return True
    return (False, "mysql module could not be loaded")


def _get_mysql_error():
    """
    Look in module context for a MySQL error. Eventually we should make a less
    ugly way of doing this.
    """
    return sys.modules[__salt__["test.ping"].__module__].__context__.pop(
        "mysql.error", None
    )


def present(name, character_set=None, collate=None, **connection_args):
    """
    Ensure that the named database is present with the specified properties

    name
        The name of the database to manage
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Database {name} is already present",
    }
    # check if database exists
    existing = __salt__["mysql.db_get"](name, **connection_args)
    if existing:
        alter_charset = False
        alter_collate = False
        existing_charset = bytes(str(existing.get("character_set")).encode()).decode()
        if character_set and character_set != existing_charset:
            alter_charset = True
            log.debug(
                "character set differes from %s : %s",
                character_set,
                existing_charset,
            )

            comment = "Database character set {} != {} needs to be updated".format(
                character_set, existing_charset
            )
            if __opts__.get("test", False):
                ret["result"] = None
                ret["comment"] = comment
            else:
                ret["comment"] = comment

        existing_collate = bytes(str(existing.get("collate")).encode()).decode()
        if collate and collate != existing_collate:
            alter_collate = True
            log.debug(
                "collate set differs from %s : %s",
                collate,
                existing_collate,
            )

            comment = "Database collate {} != {} needs to be updated".format(
                collate, existing_collate
            )
            if __opts__.get("test", False):
                ret["result"] = None
                ret["comment"] += f"\n{comment}"
                return ret
            else:
                ret["comment"] += f"\n{comment}"

        if alter_charset or alter_collate:
            if __opts__.get("test", False):
                ret["comment"] += f"\nDatabase {name} is going to be updated"
            else:
                __salt__["mysql.alter_db"](
                    name,
                    character_set=character_set,
                    collate=collate,
                    **connection_args,
                )

        current = __salt__["mysql.db_get"](name, **connection_args)
        if existing.get("collate", None) != current.get("collate", None):
            ret["changes"].update(
                {
                    "collate": {
                        "before": existing.get("collate", None),
                        "now": current.get("collate", None),
                    }
                }
            )
        if existing.get("character_set", None) != current.get("character_set", None):
            ret["changes"].update(
                {
                    "character_set": {
                        "before": existing.get("character_set", None),
                        "now": current.get("character_set", None),
                    }
                }
            )
        return ret
    else:
        err = _get_mysql_error()
        if err is not None:
            ret["comment"] = err
            ret["result"] = False
            return ret

    if __opts__.get("test", False):
        ret["result"] = None
        ret["comment"] = "Database {} is not present and needs to be created".format(
            name
        )
        return ret

    # The database is not present, make it!
    if __salt__["mysql.db_create"](
        name, character_set=character_set, collate=collate, **connection_args
    ):
        ret["comment"] = f"The database {name} has been created"
        ret["changes"][name] = "Present"
    else:
        ret["comment"] = f"Failed to create database {name}"
        err = _get_mysql_error()
        if err is not None:
            ret["comment"] += f" ({err})"
        ret["result"] = False

    return ret


def absent(name, **connection_args):
    """
    Ensure that the named database is absent

    name
        The name of the database to remove
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # check if db exists and remove it
    if __salt__["mysql.db_exists"](name, **connection_args):
        if __opts__.get("test", False):
            ret["result"] = None
            ret["comment"] = "Database {} is present and needs to be removed".format(
                name
            )
            return ret
        if __salt__["mysql.db_remove"](name, **connection_args):
            ret["comment"] = f"Database {name} has been removed"
            ret["changes"][name] = "Absent"
            return ret
        else:
            err = _get_mysql_error()
            if err is not None:
                ret["comment"] = f"Unable to remove database {name} ({err})"
                ret["result"] = False
                return ret
    else:
        err = _get_mysql_error()
        if err is not None:
            ret["comment"] = err
            ret["result"] = False
            return ret

    # fallback
    ret["comment"] = f"Database {name} is not present, so it cannot be removed"
    return ret



import pytest

import salt.proxy.restconf as restconf
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {restconf: {}}


@pytest.fixture
def patch_conn_args():
    with patch.dict(
        restconf.restconf_device,
        {
            "conn_args": {
                "hostname": None,
                "transport": None,
                "verify": None,
                "username": None,
                "password": None,
            }
        },
    ):
        yield


@pytest.fixture
def fake_query():
    with patch("salt.utils.http.query", autospec=True) as fake_query:
        yield fake_query


def test_result_of_request_should_be_returned(patch_conn_args, fake_query):
    expected_result = object()
    fake_query.return_value = expected_result
    result = restconf.request("https://example.com")
    assert result is expected_result


def test_if_None_is_provided_as_dict_payload_then_empty_string_should_be_provided_instead(
    patch_conn_args, fake_query
):
    expected_data = ""
    restconf.request("https://example.com", dict_payload=None)
    call = fake_query.mock_calls[0]

    assert call.kwargs["data"] == expected_data


def test_if_text_is_provided_as_dict_payload_then_provided_string_should_be_used(
    patch_conn_args, fake_query
):
    expected_data = "this is my fake payload"
    restconf.request("https://example.com", dict_payload=expected_data)
    call = fake_query.mock_calls[0]

    assert call.kwargs["data"] == expected_data


def test_if_dict_is_provided_as_dict_payload_then_json_text_should_be_provided(
    patch_conn_args, fake_query
):
    dict_payload = {"fnord": "something", "cool": "beans"}
    expected_data = '{"fnord": "something", "cool": "beans"}'
    restconf.request("https://example.com", dict_payload=dict_payload)
    call = fake_query.mock_calls[0]
    import json

    actual_data = json.loads(call.kwargs["data"])

    assert call.kwargs["data"] == expected_data
    assert actual_data == dict_payload


def test_if_proxy_def_connectiontest_passes_correctly(patch_conn_args, fake_query):
    fake_query.return_value = """{'body': '{\n    "ietf-restconf:yang-library-version": "2016-06-21"\n}', 'status': 200, 'dict': {'ietf-restconf:yang-library-version': '2016-06-21'}}"""
    result = restconf.connection_test()
    assert result[0] is True
    assert "yang-library-version" in result[1]


def test_if_proxy_def_connectiontest_fails_correctly(patch_conn_args, fake_query):
    fake_query.return_value = """{'body': 'fnord', 'status': 200 }"""
    result = restconf.connection_test()
    assert result[0] is False
    assert "yang-library-version" not in result[1]



import asyncio
import logging
import time
from concurrent.futures.thread import ThreadPoolExecutor

import pytest
from saltfactories.utils import random_string

import salt.channel.server
import salt.master
from tests.support.pytest.transport import PubServerChannelProcess

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
    pytest.mark.skip_on_spawning_platform(
        reason="These tests are currently broken on spawning platforms. Need to be rewritten.",
    ),
    pytest.mark.skipif(
        "grains['osfinger'] == 'Rocky Linux-8' and grains['osarch'] == 'aarch64'",
        reason="Temporarily skip on Rocky Linux 8 Arm64",
    ),
]


@pytest.fixture(scope="module", params=["zeromq", "tcp"])
def transport(request):
    yield request.param


@pytest.fixture(scope="module")
def salt_master(salt_factories, transport):
    config_defaults = {
        "transport": transport,
        "auto_accept": True,
        "sign_pub_messages": False,
    }
    factory = salt_factories.salt_master_daemon(
        random_string("ipc-master-"), defaults=config_defaults
    )
    return factory


@pytest.fixture(scope="module")
def salt_minion(salt_master):
    config_defaults = {
        "transport": salt_master.config["transport"],
        "master_ip": "127.0.0.1",
        "master_port": salt_master.config["ret_port"],
        "auth_timeout": 5,
        "auth_tries": 1,
        "master_uri": "tcp://127.0.0.1:{}".format(salt_master.config["ret_port"]),
    }
    factory = salt_master.salt_minion_daemon(
        random_string("zeromq-minion-"), defaults=config_defaults
    )
    return factory


@pytest.mark.skip_on_windows
@pytest.mark.slow_test
def test_publish_to_pubserv_ipc(salt_master, salt_minion, transport):
    """
    Test sending 10K messags to ZeroMQPubServerChannel using IPC transport

    ZMQ's ipc transport not supported on Windows
    """

    opts = dict(
        salt_master.config.copy(),
        ipc_mode="ipc",
        pub_hwm=0,
        transport=transport,
    )
    minion_opts = dict(salt_minion.config.copy(), transport=transport)
    with PubServerChannelProcess(opts, minion_opts) as server_channel:
        send_num = 10000
        expect = []
        for idx in range(send_num):
            expect.append(idx)
            load = {"tgt_type": "glob", "tgt": "*", "jid": idx}
            server_channel.publish(load)
    results = server_channel.collector.results
    assert len(results) == send_num, "{} != {}, difference: {}".format(
        len(results), send_num, set(expect).difference(results)
    )


@pytest.mark.skip_on_freebsd
@pytest.mark.slow_test
def test_issue_36469_tcp(salt_master, salt_minion, transport):
    """
    Test sending both large and small messags to publisher using TCP

    https://github.com/saltstack/salt/issues/36469
    """
    if transport == "tcp":
        pytest.skip("Test not applicable to the ZeroMQ transport.")

    def _send_small(opts, sid, num=10):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        server_channel = salt.channel.server.PubServerChannel.factory(opts)

        async def send():
            for idx in range(num):
                load = {
                    "tgt_type": "glob",
                    "tgt": "*",
                    "jid": f"{sid}-s{idx}",
                }
                await server_channel.publish(load)

        asyncio.run(send())
        # Allow some time for sends to finish
        time.sleep(3)
        server_channel.close()
        loop.close()
        asyncio.set_event_loop(None)

    def _send_large(opts, sid, num=10, size=250000 * 3):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        server_channel = salt.channel.server.PubServerChannel.factory(opts)

        async def send():
            for idx in range(num):
                load = {
                    "tgt_type": "glob",
                    "tgt": "*",
                    "jid": f"{sid}-l{idx}",
                    "xdata": "0" * size,
                }
                await server_channel.publish(load)

        asyncio.run(send())
        time.sleep(0.3)
        server_channel.close()
        loop.close()
        asyncio.set_event_loop(None)

    opts = dict(salt_master.config.copy(), ipc_mode="tcp", pub_hwm=0)
    send_num = 10 * 4
    expect = []
    with PubServerChannelProcess(opts, salt_minion.config.copy()) as server_channel:
        assert "aes" in salt.master.SMaster.secrets
        with ThreadPoolExecutor(max_workers=4) as executor:
            executor.submit(_send_small, opts, 1)
            executor.submit(_send_large, opts, 2)
            executor.submit(_send_small, opts, 3)
            executor.submit(_send_large, opts, 4)
        expect.extend([f"{a}-s{b}" for a in range(10) for b in (1, 3)])
        expect.extend([f"{a}-l{b}" for a in range(10) for b in (2, 4)])
    results = server_channel.collector.results
    assert len(results) == send_num, "{} != {}, difference: {}".format(
        len(results), send_num, set(expect).difference(results)
    )



import os

import pytest


@pytest.fixture(scope="module")
def install():
    pytest.helpers.clean_env()
    # Create an existing config
    pytest.helpers.existing_config()
    args = ["/S"]
    pytest.helpers.install_salt(args)
    yield args
    pytest.helpers.clean_env()


def test_binaries_present(install):
    # This will show the contents of the directory on failure
    inst_dir = pytest.INST_DIR
    inst_dir_exists = os.path.exists(inst_dir)
    dir_contents = os.listdir(inst_dir)
    assert os.path.exists(rf"{inst_dir}\ssm.exe")


def test_config_present(install):
    assert os.path.exists(rf"{pytest.DATA_DIR}\conf\minion")


def test_config_correct(install):
    # The config file should be the existing config, unchanged
    expected = pytest.EXISTING_CONTENT

    with open(rf"{pytest.DATA_DIR}\conf\minion") as f:
        result = f.readlines()

    assert result == expected



import pytest

import salt.exceptions
import salt.states.saltmod as saltmod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        saltmod: {
            "__env__": "base",
            "__opts__": minion_opts,
        },
    }


def test_runners():
    name = "runner-name"
    runner_1 = "runner-1"
    runner_1_ret = {
        "jid": "20170406104341210934",
        "retcode": 0,
        "ret": {
            "test_|-notify_me_|-this is a name 1_|-show_notification": {
                "comment": f"Notify me: {runner_1}",
                "name": "this is a name 1",
                "start_time": "10:43:41.487565",
                "result": True,
                "duration": 0.35,
                "__run_num__": 0,
                "__sls__": "demo",
                "changes": {},
                "__id__": "notify_me",
            }
        },
        "failed": True,
        "out": "highstate",
    }

    expected = {
        "name": name,
        "changes": {
            "ret": {
                runner_1: runner_1_ret,
            }
        },
        "result": True,
        "comment": "All runner functions executed successfully.",
    }
    mock = MagicMock(side_effect=[{"return": runner_1_ret}])
    with patch.dict(saltmod.__salt__, {"saltutil.runner": mock}):
        ret = saltmod.parallel_runners(name, runner_1)
        assert ret == expected

    runner_2 = "runner-2"
    runner_2_ret = {
        "jid": "20170406104341210934",
        "retcode": 0,
        "ret": {
            "test_|-notify_me_|-this is a name 2_|-show_notification": {
                "comment": f"Notify me: {runner_2}",
                "name": "this is a name 2",
                "start_time": "10:43:41.487565",
                "result": True,
                "duration": 0.35,
                "__run_num__": 0,
                "__sls__": "demo",
                "changes": {},
                "__id__": "notify_me",
            }
        },
        "failed": True,
        "out": "highstate",
    }
    expected["changes"]["ret"][runner_2] = runner_2_ret
    mock = MagicMock(side_effect=[{"return": runner_1_ret}, {"return": runner_2_ret}])
    with patch.dict(saltmod.__salt__, {"saltutil.runner": mock}):
        ret = saltmod.parallel_runners(
            name, {runner_1: {"name": name}, runner_2: {"name": name}}
        )
        assert ret == expected

    expected = {
        "name": name,
        "result": False,
        "changes": {},
        "comment": "The runners parameter must be a string or dict.",
    }
    ret = saltmod.parallel_runners(name, [runner_1, runner_2])
    assert ret == expected


def test_exception():
    name = "runner-name"
    runner_1 = "runner-1"
    runner_1_ret = {
        "jid": "20170406104341210934",
        "retcode": 0,
        "ret": {
            "test_|-notify_me_|-this is a name_|-show_notification": {
                "comment": "Notify me",
                "name": "this is a name",
                "start_time": "10:43:41.487565",
                "result": True,
                "duration": 0.35,
                "__run_num__": 0,
                "__sls__": "demo",
                "changes": {},
                "__id__": "notify_me",
            }
        },
        "failed": True,
        "out": "highstate",
    }

    expected = {
        "name": name,
        "result": False,
        "changes": {},
        "comment": "One of the runners raised an exception: An Exception!",
        "success": False,
    }
    runner_2 = "runner-2"
    mock = MagicMock(
        side_effect=[
            {"return": runner_1_ret},
            salt.exceptions.SaltException("An Exception!"),
        ]
    )
    with patch.dict(saltmod.__salt__, {"saltutil.runner": mock}):
        ret = saltmod.parallel_runners(
            name, {runner_1: {"name": name}, runner_2: {"name": name}}
        )
        assert ret == expected


def test_failed():

    name = "runner-name"
    runner_1 = "runner-1"
    runner_1_ret = {
        "jid": "20170406104341210934",
        "retcode": 0,
        "ret": {
            "test_|-notify_me_|-this is a name 1_|-show_notification": {
                "comment": f"Notify me: {runner_1}",
                "name": "this is a name 1",
                "start_time": "10:43:41.487565",
                "result": True,
                "duration": 0.35,
                "__run_num__": 0,
                "__sls__": "demo",
                "changes": {"foo": "bar"},
                "__id__": "notify_me",
            }
        },
        "failed": True,
        "out": "highstate",
        "exit_code": 1,
    }
    runner_2 = "runner-2"
    runner_2_ret = {
        "jid": "20170406104341210934",
        "retcode": 1,
        "ret": {
            "test_|-notify_me_|-this is a name 2_|-show_notification": {
                "comment": f"Notify me: {runner_2}",
                "name": "this is a name 2",
                "start_time": "10:43:41.487565",
                "result": False,
                "duration": 0.35,
                "__run_num__": 0,
                "__sls__": "demo",
                "changes": {},
                "__id__": "notify_me",
            }
        },
        "failed": True,
        "out": "highstate",
        "exit_code": 0,
    }

    expected = {
        "name": name,
        "changes": {
            "ret": {
                runner_1: runner_1_ret,
            }
        },
        "result": False,
        "comment": f"Runner {runner_1} failed.",
    }
    mock = MagicMock(side_effect=[{"return": runner_1_ret}])
    with patch.dict(saltmod.__salt__, {"saltutil.runner": mock}):
        ret = saltmod.parallel_runners(name, runner_1)
        assert ret == expected

    expected["changes"]["ret"][runner_2] = runner_2_ret
    mock = MagicMock(side_effect=[{"return": runner_1_ret}, {"return": runner_2_ret}])
    with patch.dict(saltmod.__salt__, {"saltutil.runner": mock}):
        ret = saltmod.parallel_runners(
            name, {runner_1: {"name": name}, runner_2: {"name": name}}
        )
        assert ret == expected

    runner_3 = "runner-3"
    runner_3_ret = {
        "jid": "20170406104341210934",
        "retcode": 1,
        "ret": {
            "test_|-notify_me_|-this is a name 2_|-show_notification": {
                "comment": f"Notify me: {runner_2}",
                "name": "this is a name 2",
                "start_time": "10:43:41.487565",
                "result": False,
                "duration": 0.35,
                "__run_num__": 0,
                "__sls__": "demo",
                "changes": {},
                "__id__": "notify_me",
            }
        },
        "failed": True,
        "out": "highstate",
        "exit_code": 1,
    }

    expected["changes"]["ret"][runner_3] = runner_3_ret
    expected.pop("comment")
    mock = MagicMock(
        side_effect=[
            {"return": runner_1_ret},
            {"return": runner_2_ret},
            {"return": runner_3_ret},
        ]
    )
    with patch.dict(saltmod.__salt__, {"saltutil.runner": mock}):
        ret = saltmod.parallel_runners(
            name,
            {
                runner_1: {"name": name},
                runner_2: {"name": name},
                runner_3: {"name": name},
            },
        )
        ret_comment = ret.pop("comment")
        assert ret == expected
        assert "Runners " in ret_comment
        assert " failed." in ret_comment
        assert runner_1 in ret_comment
        assert runner_3 in ret_comment
        assert runner_2 not in ret_comment



import pytest

import salt.states.saltmod as saltmod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        saltmod: {
            "__opts__": minion_opts,
        },
    }


def test_function():
    """
    Test to execute a single module function on a remote
    minion via salt or salt-ssh
    """
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {},
        "result": None,
        "comment": f"Function state would be executed on target {tgt}",
    }

    with patch.dict(saltmod.__opts__, {"test": True}):
        ret = saltmod.function(name, tgt)
    assert ret == expected

    expected.update(
        {
            "result": True,
            "changes": {"ret": {tgt: ""}},
            "comment": (f"Function ran successfully. Function state ran on {tgt}."),
        }
    )
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock_ret = {"larry": {"ret": "", "retcode": 0, "failed": False}}
        mock_cmd = MagicMock(return_value=mock_ret)
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock_cmd}):
            ret = saltmod.function(name, tgt)
        assert ret == expected


def test_function_when_no_minions_match():
    """
    Test to execute a single module function on a remote
    minion via salt or salt-ssh
    """
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "No minions responded",
    }

    with patch.dict(saltmod.__opts__, {"test": False}):
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": MagicMock(return_value={})}):
            ret = saltmod.function(name, tgt)
            assert ret == expected


def test_function_ssh():
    """
    Test saltmod function passes roster to saltutil.cmd
    """
    cmd_mock = MagicMock()
    with patch.dict(saltmod.__opts__, {"test": False}), patch.dict(
        saltmod.__salt__, {"saltutil.cmd": cmd_mock}
    ):
        saltmod.function("state", tgt="*", ssh=True, roster="my_roster")
    assert "roster" in cmd_mock.call_args.kwargs
    assert cmd_mock.call_args.kwargs["roster"] == "my_roster"


def test_arg():
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {"ret": {tgt: ""}},
        "result": True,
        "comment": f"Function ran successfully. Function state ran on {tgt}.",
        "warnings": ["Please specify 'arg' as a list of arguments."],
    }

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        args = ["foo", "bar"]
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, arg=" ".join(args))
            assert ret == expected
        mock.assert_called_once()
        assert "arg" in mock.call_args.kwargs
        assert mock.call_args.kwargs["arg"] == args

    expected.pop("warnings")
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        args = ["foo", "bar"]
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, arg=args)
            assert ret == expected
        mock.assert_called_once()
        assert "arg" in mock.call_args.kwargs
        assert mock.call_args.kwargs["arg"] == args


def test_batch():
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {"ret": {tgt: ""}},
        "result": True,
        "comment": f"Function ran successfully. Function state ran on {tgt}.",
    }

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        batch = "yes"
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, batch=batch)
            assert ret == expected
        mock.assert_called_once()
        assert "batch" in mock.call_args.kwargs
        assert mock.call_args.kwargs["batch"] == batch

        batch = ["yes", "no"]
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, batch=batch)
            assert ret == expected
        assert "batch" in mock.call_args.kwargs
        assert mock.call_args.kwargs["batch"] == str(batch)


def test_subset():
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {"ret": {tgt: ""}},
        "result": True,
        "comment": f"Function ran successfully. Function state ran on {tgt}.",
    }

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        subset = "yes"
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, subset=subset)
            assert ret == expected
        mock.assert_called_once()
        assert "subset" in mock.call_args.kwargs
        assert mock.call_args.kwargs["subset"] == subset


def test_ret_config():
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {"ret": {tgt: ""}},
        "result": True,
        "comment": f"Function ran successfully. Function state ran on {tgt}.",
    }

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        ret_config = {"yes": "no"}
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, ret_config=ret_config)
            assert ret == expected
        mock.assert_called_once()
        assert "ret_config" in mock.call_args.kwargs
        assert mock.call_args.kwargs["ret_config"] == ret_config


def test_ret_kwargs():
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {"ret": {tgt: ""}},
        "result": True,
        "comment": f"Function ran successfully. Function state ran on {tgt}.",
    }

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        ret_kwargs = {"yes": "no"}
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, ret_kwargs=ret_kwargs)
            assert ret == expected
        mock.assert_called_once()
        assert "ret_kwargs" in mock.call_args.kwargs
        assert mock.call_args.kwargs["ret_kwargs"] == ret_kwargs


def test_failhard():
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {"ret": {tgt: ""}},
        "result": True,
        "comment": f"Function ran successfully. Function state ran on {tgt}.",
    }

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, failhard=True)
            assert ret == expected
        mock.assert_called_once()
        assert "failhard" in mock.call_args.kwargs
        assert mock.call_args.kwargs["failhard"] is True

    with patch.dict(saltmod.__opts__, {"test": False, "failhard": True}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt)
            assert ret == expected
        mock.assert_called_once()
        assert "failhard" in mock.call_args.kwargs
        assert mock.call_args.kwargs["failhard"] is True


def test_fail_minions():
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {
            "ret": {
                tgt: "",
                "red": "red",
                "green": "green",
                "blue": "blue",
            },
        },
        "result": True,
    }

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
                "red": {
                    "ret": "red",
                    "retcode": 0,
                    "failed": False,
                },
                "green": {
                    "ret": "green",
                    "retcode": 0,
                    "failed": False,
                },
                "blue": {
                    "ret": "blue",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, fail_minions="red")
            ret_comment = ret.pop("comment")
            assert ret == expected
            assert "Function ran successfully. Function state ran on " in ret_comment
            for part in (tgt, "red", "green", "blue"):
                assert part in ret_comment

        mock.assert_called_once()

        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, fail_minions="red,green")
            ret_comment = ret.pop("comment")
            assert ret == expected
            assert "Function ran successfully. Function state ran on " in ret_comment
            for part in (tgt, "red", "green", "blue"):
                assert part in ret_comment

        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, fail_minions=["red", "green"])
            ret_comment = ret.pop("comment")
            assert ret == expected
            assert "Function ran successfully. Function state ran on " in ret_comment
            for part in (tgt, "red", "green", "blue"):
                assert part in ret_comment

        expected["warnings"] = [
            "'fail_minions' needs to be a list or a comma separated string. Ignored."
        ]
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, fail_minions=())
            ret_comment = ret.pop("comment")
            assert ret == expected
            assert "Function ran successfully. Function state ran on " in ret_comment
            for part in (tgt, "red", "green", "blue"):
                assert part in ret_comment

    expected.pop("warnings")
    expected["changes"]["ret"]["red"] = False
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
                "red": {
                    "ret": "red",
                    "retcode": 0,
                    "failed": True,
                },
                "green": {
                    "ret": "green",
                    "retcode": 0,
                    "failed": False,
                },
                "blue": {
                    "ret": "blue",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, fail_minions="red")
            ret_comment = ret.pop("comment")
            assert ret == expected
            assert "Function ran successfully. Function state ran on " in ret_comment
            for part in (tgt, "red", "green", "blue"):
                assert part in ret_comment

        mock.assert_called_once()

    expected["result"] = False
    expected["changes"]["ret"]["green"] = False
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
                "red": {
                    "ret": "red",
                    "retcode": 0,
                    "failed": True,
                },
                "green": {
                    "ret": "green",
                    "retcode": 0,
                    "failed": True,
                },
                "blue": {
                    "ret": "blue",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, fail_minions="red")
            ret_comment = ret.pop("comment")
            assert ret == expected
            assert "Running function state failed on minions: green " in ret_comment
            assert "Function state ran on " in ret_comment
            for part in (tgt, "red", "green", "blue"):
                assert part in ret_comment

        mock.assert_called_once()

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                tgt: {
                    "ret": "",
                    "retcode": 0,
                    "failed": False,
                },
                "red": {
                    "ret": "red",
                    "retcode": 1,
                    "failed": True,
                },
                "green": {
                    "ret": "green",
                    "retcode": 1,
                    "failed": True,
                },
                "blue": {
                    "ret": "blue",
                    "retcode": 0,
                    "failed": False,
                },
            },
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, fail_minions="red")
            ret_comment = ret.pop("comment")
            assert ret == expected
            try:
                assert (
                    "Running function state failed on minions: green, red"
                    in ret_comment
                )
            except AssertionError:
                assert (
                    "Running function state failed on minions: red, green"
                    in ret_comment
                )
            assert "Function state ran on " in ret_comment
            for part in (tgt, "red", "green", "blue"):
                assert part in ret_comment

        mock.assert_called_once()


def test_exception_raised():
    name = "state"
    tgt = "larry"

    expected = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "I'm an exception!",
    }

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(side_effect=Exception("I'm an exception!"))
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.function(name, tgt, failhard=True)
            assert ret == expected
        mock.assert_called_once()
        assert "failhard" in mock.call_args.kwargs
        assert mock.call_args.kwargs["failhard"] is True
